# Copyright (C) 2009-2014 Yann Leboulanger <asterix AT lagaule.org>
#
# This file is part of Gajim.
#
# SPDX-License-Identifier: GPL-3.0-only

from __future__ import annotations

import textwrap
from collections.abc import Iterator
from urllib.parse import quote

from gi.repository import Gio
from gi.repository import GLib
from gi.repository.Gio import Menu
from nbxmpp import JID

from gajim.common import app
from gajim.common import types
from gajim.common.const import XmppUriQuery
from gajim.common.i18n import _
from gajim.common.i18n import get_short_lang_code
from gajim.common.i18n import p_
from gajim.common.modules.contacts import BareContact
from gajim.common.modules.contacts import can_add_to_roster
from gajim.common.modules.contacts import GroupchatContact
from gajim.common.modules.contacts import GroupchatOfflineParticipant
from gajim.common.modules.contacts import GroupchatParticipant
from gajim.common.storage.archive.const import ChatDirection
from gajim.common.storage.archive.models import Message
from gajim.common.structs import VariantMixin
from gajim.common.util.muc import is_affiliation_change_allowed
from gajim.common.util.muc import is_moderation_allowed
from gajim.common.util.muc import is_role_change_allowed
from gajim.common.util.text import escape_iri_path_segment
from gajim.common.util.uri import MailUri
from gajim.common.util.uri import UriT
from gajim.common.util.uri import XmppIri

from gajim.gtk.const import MuteState
from gajim.gtk.structs import AccountJidParam
from gajim.gtk.structs import AddChatActionParams
from gajim.gtk.structs import ChatListEntryParam
from gajim.gtk.structs import DeleteMessageParam
from gajim.gtk.structs import ExportHistoryParam
from gajim.gtk.structs import ModerateAllMessagesParam
from gajim.gtk.structs import ModerateMessageParam
from gajim.gtk.structs import MuteContactParam
from gajim.gtk.structs import OccupantParam
from gajim.gtk.structs import RetractMessageParam
from gajim.gtk.util.misc import is_message_correctable

MenuValueT = None | str | GLib.Variant | VariantMixin
MenuItemListT = list[tuple[str, str, MenuValueT]]
UriMenuItemsT = list[tuple[str, list[str] | VariantMixin, str]]


class GajimMenu(Gio.Menu):
    def __init__(self):
        Gio.Menu.__init__(self)

    @classmethod
    def from_list(cls, menulist: MenuItemListT) -> GajimMenu:
        menu = cls()
        for item in menulist:
            menuitem = make_menu_item(*item)
            menu.append_item(menuitem)
        return menu

    def add_item(
        self, label: str, action: str, value: MenuValueT | None = None
    ) -> None:
        item = make_menu_item(label, action, value)
        self.append_item(item)

    def add_submenu(self, label: str) -> GajimMenu:
        menu = GajimMenu()
        self.append_submenu(label, menu)
        return menu


def make_menu_item(
    label: str, action: str | None = None, value: MenuValueT = None
) -> Gio.MenuItem:
    item = Gio.MenuItem.new(label)

    if value is None:
        item.set_action_and_target_value(action, None)
        return item

    item = Gio.MenuItem.new(label)
    if isinstance(value, str):
        item.set_action_and_target_value(action, GLib.Variant("s", value))
    elif isinstance(value, VariantMixin):
        item.set_action_and_target_value(action, value.to_variant())
    else:
        item.set_action_and_target_value(action, value)
    return item


def get_self_contact_menu(contact: types.BareContact) -> GajimMenu:
    account = contact.account
    jid = contact.jid

    menu = GajimMenu()

    menu.add_item(_("Profile"), f"app.{account}-profile", account)
    submenu = get_send_file_submenu()
    menu.append_submenu(_("Send File"), submenu)

    params = ExportHistoryParam(account=account, jid=jid)
    menu.add_item(_("Export History…"), "app.export-history", params)

    params = AccountJidParam(account=account, jid=jid)
    menu.add_item(_("Remove History…"), "app.remove-history", params)
    return menu


def get_singlechat_menu(contact: types.BareContact) -> GajimMenu:
    account = contact.account

    params = AccountJidParam(account=account, jid=contact.jid)
    export_params = ExportHistoryParam(account=account, jid=contact.jid)

    menu = GajimMenu()

    submenu = get_send_file_submenu()
    menu.append_submenu(_("Send File"), submenu)
    menu.add_item(_("Block Contact…"), f"app.{account}-block-contact", params)

    # Disable because not maintained
    # menu.add_item(_("Start Voice Call…"), "win.start-voice-call")
    # menu.add_item(_("Start Video Call…"), "win.start-video-call")

    if can_add_to_roster(contact):
        menu.add_item(_("Add Contact…"), f"app.{account}-add-contact", params)

    if contact.is_in_roster:
        roster_add_params = GLib.Variant("as", [account, str(contact.jid)])
        menu.add_item(
            _("Remove Contact…"), f"app.{account}-remove-contact", roster_add_params
        )

    jids = [str(c.jid) for c in contact.get_resources()]
    if not jids:
        jids = [str(contact.jid)]

    menu.add_item(
        _("Execute Command…"),
        f"app.{account}-execute-command",
        GLib.Variant("(sas)", (account, jids)),
    )

    menu.add_item(_("Export History…"), "app.export-history", export_params)
    menu.add_item(_("Remove History…"), "app.remove-history", params)

    return menu


def get_private_chat_menu(contact: types.GroupchatParticipant) -> GajimMenu:
    menu = GajimMenu()

    value = GLib.Variant("as", [""])

    menu.add_item(_("Upload File…"), "win.send-file-httpupload", value)

    params = ExportHistoryParam(account=contact.account, jid=contact.jid)
    menu.add_item(_("Export History…"), "app.export-history", params)

    params = AccountJidParam(account=contact.account, jid=contact.jid)
    menu.add_item(_("Remove History…"), "app.remove-history", params)

    occupant_param = None
    if contact.occupant_id is not None:
        occupant_param = OccupantParam(
            contact.account, contact.room.jid, contact.name, contact.occupant_id
        )

    if contact.is_blocked:
        menu.add_item(_("Unblock"), "win.muc-user-unblock", occupant_param)
    else:
        menu.add_item(_("Block…"), "win.muc-user-block", occupant_param)

    real_contact = contact.get_real_contact()
    if real_contact is not None and can_add_to_roster(real_contact):
        params = AccountJidParam(account=contact.account, jid=real_contact.jid)
        menu.add_item(_("Add Contact…"), f"app.{contact.account}-add-contact", params)

    return menu


def get_send_file_submenu() -> GajimMenu:
    menu = GajimMenu()

    value = GLib.Variant("as", [""])

    menu.add_item(_("Upload File…"), "win.send-file-httpupload", value)
    # menu.add_item(_('Send File Directly…'), 'win.send-file-jingle', value)
    return menu


def get_groupchat_menu(contact: GroupchatContact) -> GajimMenu:
    menu = GajimMenu()

    menu.add_item(_("Change Nickname…"), "win.muc-change-nickname", None)
    menu.add_item(_("Request Voice"), "win.muc-request-voice", None)
    menu.add_item(_("Execute Command…"), "win.muc-execute-command", "")

    params = ExportHistoryParam(account=contact.account, jid=contact.jid)
    menu.add_item(_("Export History…"), "app.export-history", params)

    params = AccountJidParam(account=contact.account, jid=contact.jid)
    menu.add_item(_("Remove History…"), "app.remove-history", params)

    return menu


def get_account_menu(account: str) -> GajimMenu:
    client = app.get_client(account)
    server_jid = client.get_own_jid().domain
    assert server_jid is not None

    params = GLib.Variant("(sas)", (account, [server_jid]))

    menuitems: MenuItemListT = [
        (_("Discover Services…"), f"app.{account}-services", account),
        (_("Execute Command…"), f"app.{account}-execute-command", params),
        (_("PEP Configuration"), f"app.{account}-pep-config", account),
    ]

    return GajimMenu.from_list(menuitems)


def get_encryption_menu() -> GajimMenu:
    menuitems: MenuItemListT = [
        (_("Disabled"), "win.set-encryption", ""),
        ("OMEMO", "win.set-encryption", "OMEMO"),
        ("OpenPGP", "win.set-encryption", "OpenPGP"),
        ("PGP", "win.set-encryption", "PGP"),
    ]

    return GajimMenu.from_list(menuitems)


def get_banner_narrow_menu(
    account: str, jid: JID | None, base_menu: Gio.Menu
) -> Menu | None:
    if jid is None:
        return None

    client = app.get_client(account)
    contact = client.get_module("Contacts").get_contact(jid)

    menuitems: MenuItemListT = [
        (_("Chat Details and Settings"), "win.show-contact-info", None)
    ]

    if isinstance(contact, GroupchatContact):
        menuitems.append((_("Invite to this chat…"), "win.muc-invite", None))
    if not contact.is_pm_contact:
        menuitems.append((_("Share…"), "win.show-chat-share", None))

    combined = GajimMenu.from_list(menuitems)
    combined.append_section(None, base_menu)
    return combined


def get_message_input_extra_context_menu() -> Gio.Menu:
    menuitems: MenuItemListT = [
        (_("Clear"), "win.input-clear", None),
        (_("Paste as Code Block"), "win.input-paste-as-code-block", None),
        (_("Paste as Quote"), "win.input-paste-as-quote", None),
    ]

    menu = GajimMenu.from_list(menuitems)
    extra_menu = Gio.Menu()
    extra_menu.append_section(None, menu)
    return extra_menu


def get_conv_action_context_menu(account: str, selected_text: str) -> Gio.Menu:
    menuitems: MenuItemListT = []

    selected_text_short = textwrap.shorten(selected_text, width=10, placeholder="…")

    uri_text = quote(selected_text.encode("utf-8"))

    # Wikipedia search
    if app.settings.get("always_english_wikipedia"):
        uri = f"https://en.wikipedia.org/wiki/Special:Search?search={uri_text}"
    else:
        uri = (
            f"https://{get_short_lang_code()}.wikipedia.org/"
            f"wiki/Special:Search?search={uri_text}"
        )

    menuitems.append(
        (_("Read _Wikipedia Article"), "app.open-link", GLib.Variant("s", uri))
    )

    # Dictionary search
    dictionary_title = _("Look it up in _Dictionary")
    dict_link = app.settings.get("dictionary_url")
    if dict_link == "WIKTIONARY":
        # Default is wikitionary.org
        if app.settings.get("always_english_wiktionary"):
            uri = f"https://en.wiktionary.org/wiki/Special:Search?search={uri_text}"
        else:
            uri = (
                f"https://{get_short_lang_code()}.wiktionary.org/"
                f"wiki/Special:Search?search={uri_text}"
            )
    else:
        if dict_link.find("%s") == -1:
            # There has to be a '%s' in the url if it’s not WIKTIONARY
            dictionary_title = _('Dictionary URL is missing a "%s"')
        else:
            uri = dict_link % uri_text

    menuitems.append((dictionary_title, "app.open-link", GLib.Variant("s", uri)))

    # Generic search
    search_link = app.settings.get("search_engine")
    if search_link.find("%s") == -1:
        # There has to be a '%s' in the url
        search_title = _('Web Search URL is missing a "%s"')
    else:
        search_title = _("Web _Search for it")
        uri = search_link % uri_text

    menuitems.append((search_title, "app.open-link", GLib.Variant("s", uri)))

    # Open as URI
    variant = GLib.Variant("s", f"https://{uri_text}")
    menuitems.append((_("Open as _Link"), "app.open-link", variant))

    menu = GajimMenu.from_list(menuitems)
    extra_menu = Gio.Menu()
    extra_menu.append_section(
        _('Actions for "%s"') % escape_mnemonic(selected_text_short), menu
    )
    return extra_menu


def _xmpp_uri_context_menu(uri: XmppIri, account: str) -> UriMenuItemsT:
    menu_items: UriMenuItemsT = [
        ("copy-text", [str(uri.jid)], _("Copy XMPP Address")),
    ]
    query_type = XmppUriQuery.from_str_or_none(uri.action)
    if query_type in (XmppUriQuery.NONE, XmppUriQuery.MESSAGE):
        params = AccountJidParam(account=account, jid=uri.jid)
        menu_items.extend(
            [
                ("open-chat", [account, str(uri.jid)], _("Start Chat…")),
                (f"{account}-add-contact", params, _("Add Contact…")),
            ]
        )
        return menu_items

    if query_type == XmppUriQuery.JOIN:
        menu_items.append(
            (
                "open-chat",
                [account, str(uri.jid)],
                _("Join Groupchat…"),
            )
        )

    return menu_items


def _ambiguous_addr_context_menu(uri: MailUri, account: str) -> UriMenuItemsT:
    mailto = "mailto:" + escape_iri_path_segment(uri.addr)

    # addr could be a non valid jid
    try:
        params = AccountJidParam(account=account, jid=JID.from_string(uri.addr))
    except Exception:
        params = None

    items: UriMenuItemsT = [
        ("copy-text", [uri.addr], _("Copy XMPP Address/Email")),
        ("open-link", [mailto], _("Open Email Composer")),
        ("open-chat", [account, uri.addr], _("Start Chat…")),
    ]

    if params is not None:
        items.append((f"{account}-add-contact", params, _("Add Contact…")))

    return items


def get_uri_context_menu(account: str, uri: UriT) -> Gio.Menu | None:
    match uri:
        case XmppIri():
            menu_data = _xmpp_uri_context_menu(uri, account)
        case MailUri():
            menu_data = _ambiguous_addr_context_menu(uri, account)
        case _:
            return None

    menuitems: MenuItemListT = []

    for action, args, label in menu_data:
        if isinstance(args, VariantMixin):
            value = args.to_variant()
        elif len(args) == 1:
            value = GLib.Variant.new_string(args[0])
        else:
            value = GLib.Variant.new_strv(args)

        menuitems.append(
            (label, f"app.{action}", value),
        )

    return GajimMenu.from_list(menuitems)


def get_account_notifications_menu(account: str) -> GajimMenu:
    menuitems: MenuItemListT = [
        (
            _("Deny all contact requests"),
            f"win.subscription-deny-all-{account}",
            None,
        ),
    ]
    return GajimMenu.from_list(menuitems)


def get_subscription_menu(account: str, jid: JID) -> GajimMenu:
    add_chat_params = AddChatActionParams(
        account=account, jid=jid, type="chat", select=True
    )
    account_jid_params = AccountJidParam(account=account, jid=jid)
    menuitems: MenuItemListT = [
        (_("Start Chat"), "win.add-chat", add_chat_params),
        (_("Details"), "win.chat-contact-info", account_jid_params),
        (_("Block"), f"app.{account}-block-contact", account_jid_params),
    ]

    return GajimMenu.from_list(menuitems)


def get_activity_feed_menu() -> GajimMenu:
    menuitems: MenuItemListT = [
        (_("Mark all activities as read"), "win.mark-all-activities-as-read", None),
    ]

    return GajimMenu.from_list(menuitems)


def get_start_chat_button_menu() -> GajimMenu:
    value = GLib.Variant("as", ["", ""])

    menuitems: MenuItemListT = [
        (_("Start Chat…"), "app.start-chat", value),
        (_("Create Group Chat…"), "app.create-groupchat", ""),
        (_("Add Contact…"), "app.add-contact", ""),
    ]

    return GajimMenu.from_list(menuitems)


def get_start_chat_row_menu(account: str, jid: JID | None) -> GajimMenu | None:
    if jid is None:
        return None

    client = app.get_client(account)
    contact = client.get_module("Contacts").get_contact(jid)

    menuitems: MenuItemListT = []

    command_params = GLib.Variant("(sas)", (account, [str(jid)]))
    menuitems.append(
        (_("Execute Command…"), f"app.{account}-execute-command", command_params)
    )

    if isinstance(contact, GroupchatContact):
        account_jid_params = AccountJidParam(account=account, jid=jid)
        menuitems.append(
            (_("Forget this Group Chat…"), "app.forget-groupchat", account_jid_params)
        )

    return GajimMenu.from_list(menuitems)


def get_chat_list_row_menu(
    workspace_id: str, account: str, jid: JID, pinned: bool
) -> GajimMenu:
    client = app.get_client(account)
    contact = client.get_module("Contacts").get_contact(jid)
    assert isinstance(contact, BareContact | GroupchatContact | GroupchatParticipant)

    menu = GajimMenu()

    params = ChatListEntryParam(
        workspace_id=workspace_id,
        source_workspace_id=workspace_id,
        account=account,
        jid=jid,
    )

    toggle_label = _("Unpin Chat") if pinned else _("Pin Chat")
    menu.add_item(toggle_label, "win.toggle-chat-pinned", params)

    submenu = menu.add_submenu(_("Move Chat"))
    if app.settings.get_workspace_count() > 1:
        for name, params in get_workspace_params(workspace_id, account, jid):
            submenu.add_item(name, "win.move-chat-to-workspace", params)

    params = ChatListEntryParam(
        workspace_id="", source_workspace_id=workspace_id, account=account, jid=jid
    )

    submenu.add_item(_("New Workspace"), "win.move-chat-to-workspace", params)

    if isinstance(contact, GroupchatParticipant):
        real_contact = contact.get_real_contact()
    elif isinstance(contact, BareContact):
        real_contact = contact
    else:
        real_contact = None

    if real_contact is not None and can_add_to_roster(real_contact):
        params = AccountJidParam(account=account, jid=real_contact.jid)
        menu.add_item(_("Add Contact…"), f"app.{account}-add-contact", params)

    if app.window.get_chat_unread_count(account, jid, include_silent=True):
        params = AccountJidParam(account=account, jid=jid)
        menu.add_item(_("Mark as read"), f"app.{account}-mark-as-read", params)

    if contact.is_muted:
        menu.add_item(
            _("Unmute Chat"),
            "app.mute-chat",
            MuteContactParam(account=account, jid=jid, state=MuteState.UNMUTE),
        )
    else:
        submenu = menu.add_submenu(_("Mute Chat"))
        for state, label in MuteState.iter():
            submenu.add_item(
                label,
                "app.mute-chat",
                MuteContactParam(account=account, jid=jid, state=state),
            )

    return menu


def get_workspace_params(
    current_workspace_id: str, account: str, jid: JID
) -> Iterator[tuple[str, ChatListEntryParam]]:
    for workspace_id in app.settings.get_workspaces():
        if workspace_id == current_workspace_id:
            continue
        name = app.settings.get_workspace_setting(workspace_id, "name")
        params = ChatListEntryParam(
            workspace_id=workspace_id,
            source_workspace_id=current_workspace_id,
            account=account,
            jid=jid,
        )
        yield name, params


def get_groupchat_admin_menu(
    self_contact: GroupchatParticipant,
    contact: GroupchatParticipant | GroupchatOfflineParticipant,
) -> GajimMenu:
    menu = GajimMenu()

    if contact.real_jid is None:
        menu.add_item(_("Not Available"), "dummy", None)
        return menu

    action = "win.muc-change-affiliation"
    real_jid = str(contact.real_jid)

    if is_affiliation_change_allowed(self_contact, contact, "owner"):
        value = GLib.Variant("as", [real_jid, "owner"])
        menu.add_item(_("Make Owner"), action, value)

    if is_affiliation_change_allowed(self_contact, contact, "admin"):
        value = GLib.Variant("as", [real_jid, "admin"])
        menu.add_item(_("Make Admin"), action, value)

    if is_affiliation_change_allowed(self_contact, contact, "member"):
        value = GLib.Variant("as", [real_jid, "member"])
        menu.add_item(_("Make Member"), action, value)

    if is_affiliation_change_allowed(self_contact, contact, "none"):
        value = GLib.Variant("as", [real_jid, "none"])
        menu.add_item(_("Revoke Member"), action, value)

    if is_affiliation_change_allowed(self_contact, contact, "outcast"):
        menu.add_item(_("Ban…"), "win.muc-ban", real_jid)

    if not menu.get_n_items():
        menu.add_item(_("Not Available"), "dummy", None)

    return menu


def get_groupchat_mod_menu(
    self_contact: GroupchatParticipant,
    contact: GroupchatParticipant | GroupchatOfflineParticipant,
) -> GajimMenu:
    menu = GajimMenu()

    if not contact.is_available:
        menu.add_item(_("Not Available"), "dummy", None)
        return menu

    assert isinstance(contact, GroupchatParticipant)

    contact_name = str(contact.name)

    if is_role_change_allowed(self_contact, contact):
        menu.add_item(_("Kick…"), "win.muc-kick", contact_name)

    action = "win.muc-change-role"

    if is_role_change_allowed(self_contact, contact):
        if contact.role.is_visitor:
            value = GLib.Variant("as", [contact_name, "participant"])
            menu.add_item(_("Grant Voice"), action, value)
        else:
            value = GLib.Variant("as", [contact_name, "visitor"])
            menu.add_item(_("Revoke Voice"), action, value)

    if not menu.get_n_items():
        menu.add_item(_("Not Available"), "dummy", None)

    return menu


def get_groupchat_participant_menu(
    account: str,
    self_contact: GroupchatParticipant,
    contact: GroupchatParticipant | GroupchatOfflineParticipant,
) -> GajimMenu:
    group_chat = self_contact.room
    disco = group_chat.get_disco()
    assert disco is not None
    muc_prefer_direct_msg = app.settings.get("muc_prefer_direct_msg")

    dm_params = None
    if (
        disco.muc_is_nonanonymous
        and muc_prefer_direct_msg
        and contact.real_jid is not None
    ):
        dm_params = AddChatActionParams(
            account=account, jid=contact.real_jid, type="chat", select=True
        )

    elif isinstance(contact, GroupchatParticipant):
        dm_params = AddChatActionParams(
            account=account, jid=contact.jid, type="pm", select=True
        )

    value = str(contact.name)

    general_items: MenuItemListT = [
        (_("Details"), "win.muc-contact-info", value),
        (_("Execute Command…"), "win.muc-execute-command", value),
    ]

    if dm_params is not None:
        general_items.insert(
            0,
            (_("Direct Message…"), "win.add-chat", dm_params),
        )

    occupant_param = None
    if contact.occupant_id is not None:
        occupant_param = OccupantParam(
            account, contact.room.jid, contact.name, contact.occupant_id
        )

    if contact.is_blocked:
        general_items.append((_("Unblock"), "win.muc-user-unblock", occupant_param))
    else:
        general_items.append((_("Block…"), "win.muc-user-block", occupant_param))

    real_contact = contact.get_real_contact()
    if real_contact is not None and can_add_to_roster(real_contact):
        params = AccountJidParam(account=account, jid=real_contact.jid)
        action = f"app.{account}-add-contact"
        general_items.insert(1, (_("Add Contact…"), action, params.to_variant()))

    mod_menu = get_groupchat_mod_menu(self_contact, contact)
    admin_menu = get_groupchat_admin_menu(self_contact, contact)

    menu = GajimMenu.from_list(general_items)
    menu.append_section(_("Moderation"), mod_menu)
    menu.append_section(_("Administration"), admin_menu)
    return menu


def get_component_search_menu(jid: str | None, copy_text: str) -> GajimMenu:
    menuitems: MenuItemListT = [
        (_("Copy"), "app.copy-text", copy_text),
    ]

    if jid is not None:
        menuitems.append((_("Start Chat…"), "app.start-chat", jid))

    return GajimMenu.from_list(menuitems)


def get_chat_row_menu(
    contact: types.ChatContactT,
    copy_text: str | None,
    message: Message,
    original_message: Message,
) -> GajimMenu:
    menu_items: MenuItemListT = []
    is_retracted = original_message.is_retracted()

    menu_items.append((p_("Message row action", "Copy"), "win.copy-message", copy_text))

    if original_message.id is not None:
        if is_message_correctable(original_message):
            menu_items.append(
                (p_("Message row action", "Edit…"), "win.correct-message", None)
            )

    param = None
    if not is_retracted:
        param = _get_retract_param(contact, original_message)

    menu_items.append(
        (p_("Message row action", "Retract…"), "win.retract-message", param)
    )

    single_param, multiple_param = None, None
    if not is_retracted:
        single_param, multiple_param = _get_moderate_params(contact, original_message)

    menu_items.append(
        (p_("Message row action", "Moderate…"), "win.moderate-message", single_param)
    )
    menu_items.append(
        (
            p_("Message row action", "Moderate all messages…"),
            "win.moderate-all-messages",
            multiple_param,
        )
    )
    menu_items.append(
        (
            p_("Message row action", "Select Messages…"),
            "win.activate-message-selection",
            GLib.Variant("u", message.pk),
        )
    )

    param = DeleteMessageParam(
        account=contact.account, jid=contact.jid, pk=original_message.pk
    )

    menu_items.append(
        (
            p_("Message row action", "Delete Message Locally…"),
            "win.delete-message-locally",
            param,
        )
    )

    return GajimMenu.from_list(menu_items)


def _get_retract_param(
    contact: types.ChatContactT,
    original_message: Message,
) -> RetractMessageParam | None:
    if original_message.direction == ChatDirection.INCOMING:
        return

    if isinstance(contact, GroupchatContact):
        if not contact.is_joined:
            return

    if isinstance(contact, GroupchatParticipant) and not contact.room.is_joined:
        return

    if not (retract_ids := original_message.get_ids_for_retract()):
        return

    return RetractMessageParam(contact.account, contact.jid, retract_ids)


def _get_moderate_params(
    contact: types.ChatContactT,
    original_message: Message,
) -> tuple[ModerateMessageParam | None, ModerateAllMessagesParam | None]:
    if not isinstance(contact, GroupchatContact):
        return None, None

    if not contact.is_joined:
        return None, None

    if (resource := original_message.resource) is None:
        # Message from a MUC
        return None, None

    self_contact = contact.get_self()
    assert self_contact is not None
    if not is_moderation_allowed(self_contact):
        return None, None

    disco_info = app.storage.cache.get_last_disco_info(contact.jid)
    assert disco_info is not None

    if not disco_info.has_message_moderation:
        return None, None

    stanza_ids = original_message.get_ids_for_moderate()
    if not stanza_ids:
        return None, None

    ns = disco_info.moderation_namespace
    assert ns is not None
    single_param = ModerateMessageParam(
        account=contact.account,
        jid=contact.jid,
        stanza_ids=stanza_ids,
        namespace=ns,
    )

    multiple_param = None
    if original_message.occupant is not None:
        multiple_param = ModerateAllMessagesParam(
            account=contact.account,
            jid=contact.jid,
            occupant_id=original_message.occupant.id,
            nickname=resource,
            namespace=ns,
        )

    return single_param, multiple_param


def get_preview_menu(uri: str, *, encrypted: bool = False) -> GajimMenu:
    menu_items: MenuItemListT = [
        (_("_Copy Link"), "app.copy-text", uri),
    ]

    if not encrypted:
        menu_items.append((_("Open Link in _Browser"), "app.open-link", uri))

    return GajimMenu.from_list(menu_items)


def get_format_menu() -> GajimMenu:
    menuitems: MenuItemListT = [
        (_("Bold"), "win.input-bold", None),
        (_("Italic"), "win.input-italic", None),
        (_("Strike"), "win.input-strike", None),
        (_("Code"), "win.input-preformatted", None),
        (_("Code Block"), "win.input-preformatted-block", None),
    ]

    return GajimMenu.from_list(menuitems)


def get_workspace_menu(workspace_id: str) -> GajimMenu:
    remove_action = "win.dummy"
    if app.settings.get_workspace_count() > 1:
        remove_action = "win.remove-workspace"

    menuitems: MenuItemListT = [
        (_("Mark as read"), "win.mark-workspace-as-read", workspace_id),
        (_("Edit…"), "win.edit-workspace", workspace_id),
        (_("Remove"), remove_action, workspace_id),
    ]

    return GajimMenu.from_list(menuitems)


def get_manage_roster_menu(
    account: str, groups: list[str], single_selection: bool
) -> GajimMenu:
    menu = GajimMenu()

    menuitems: MenuItemListT = [
        (_("New Group…"), f"win.{account}-move-to-new-group", None)
    ]

    for group in groups:
        menuitems.append((group, f"win.{account}-move-to-group", group))

    menu.append_submenu(_("Move to Group"), GajimMenu.from_list(menuitems))

    menuitems: MenuItemListT = [
        (_("New Group…"), f"win.{account}-add-to-new-group", None)
    ]

    for group in groups:
        menuitems.append((group, f"win.{account}-add-to-group", group))

    menu.append_submenu(_("Add to Group"), GajimMenu.from_list(menuitems))

    menu.add_item(_("Remove from Group"), f"win.{account}-remove-from-group", None)
    if single_selection:
        menu.add_item(_("Rename…"), f"win.{account}-change-name", None)

    menu.add_item(
        _("Remove from Contact List…"), f"win.{account}-remove-from-roster", None
    )

    return menu


def get_manage_roster_import_menu(
    account: str, accounts: list[tuple[str, str]]
) -> GajimMenu:
    menu = GajimMenu()
    menu.add_item(_("Import from File"), f"win.{account}-import-from-file", None)

    if accounts:
        menuitems: MenuItemListT = []
        for remote_account, label in accounts:
            menuitems.append(
                (label, f"win.{account}-import-from-account", remote_account)
            )

        menu.append_submenu(_("Import from Account"), GajimMenu.from_list(menuitems))
    else:
        menu.add_item(_("Import from Account"), "", None)

    return menu


def get_start_chat_menu():
    menuitems: MenuItemListT = [
        (_("Sort by Status"), "win.sort-by-show", None),
    ]

    return GajimMenu.from_list(menuitems)


def escape_mnemonic(label: str | None) -> str | None:
    if label is None:
        return None
    # Underscore inside a label means the next letter is a keyboard
    # shortcut. To show an underscore we have to use double underscore
    return label.replace("_", "__")



# This file is part of Gajim.
#
# SPDX-License-Identifier: GPL-3.0-only

from __future__ import annotations

from typing import Any
from typing import cast
from typing import Literal
from typing import TYPE_CHECKING

import logging
import sys

from gi.repository import Adw
from gi.repository import Gdk
from gi.repository import Gio
from gi.repository import GLib
from gi.repository import Gtk
from nbxmpp import JID

from gajim.common import app
from gajim.common import events
from gajim.common import ged
from gajim.common import types
from gajim.common.client import Client
from gajim.common.const import Direction
from gajim.common.const import Display
from gajim.common.const import SimpleClientState
from gajim.common.ged import EventHelper
from gajim.common.i18n import _
from gajim.common.modules.contacts import BareContact
from gajim.common.modules.contacts import GroupchatContact
from gajim.common.modules.contacts import GroupchatParticipant
from gajim.common.modules.contacts import ResourceContact
from gajim.common.storage.archive.const import MessageType
from gajim.common.storage.archive.models import Message
from gajim.common.util.uri import InvalidUri
from gajim.common.util.uri import XmppIri

from gajim.gtk import sound
from gajim.gtk.about import AboutDialog
from gajim.gtk.activity_list import ActivityListView
from gajim.gtk.activity_list import Reaction
from gajim.gtk.alert import AlertDialog
from gajim.gtk.alert import CancelDialogResponse
from gajim.gtk.alert import ConfirmationAlertDialog
from gajim.gtk.alert import DialogEntry
from gajim.gtk.alert import DialogResponse
from gajim.gtk.alert import InformationAlertDialog
from gajim.gtk.app_side_bar import AppSideBar
from gajim.gtk.chat_list import ChatList
from gajim.gtk.chat_list_row import ChatListRow
from gajim.gtk.chat_stack import ChatStack
from gajim.gtk.const import MAIN_WIN_ACTIONS
from gajim.gtk.emoji_chooser import EmojiChooser
from gajim.gtk.main_stack import MainStack
from gajim.gtk.preview.preview import PreviewWidget
from gajim.gtk.start_chat import parse_uri
from gajim.gtk.structs import AccountJidParam
from gajim.gtk.structs import actionmethod
from gajim.gtk.structs import AddChatActionParams
from gajim.gtk.structs import ChatListEntryParam
from gajim.gtk.structs import DeleteMessageParam
from gajim.gtk.structs import ModerateAllMessagesParam
from gajim.gtk.structs import ModerateMessageParam
from gajim.gtk.structs import OccupantParam
from gajim.gtk.structs import RetractMessageParam
from gajim.gtk.util.misc import get_ui_string
from gajim.gtk.util.window import get_app_window
from gajim.gtk.util.window import open_window
from gajim.gtk.util.window import resize_window

if TYPE_CHECKING:
    from gajim.gtk.control import ChatControl

if app.is_display(Display.X11):
    from gi.repository import GdkX11

if app.is_display(Display.WIN32):
    import gi

    gi.require_version("GdkWin32", "4.0")
    from gi.repository import GdkWin32

log = logging.getLogger("gajim.gtk.main")


@Gtk.Template(string=get_ui_string("main.ui"))
class MainWindow(Adw.ApplicationWindow, EventHelper):
    __gtype_name__ = "MainWindow"

    _header_bar: Adw.HeaderBar = Gtk.Template.Child()
    _back_button: Gtk.Button = Gtk.Template.Child()
    _split_view: Adw.OverlaySplitView = Gtk.Template.Child()
    _app_side_bar: AppSideBar = Gtk.Template.Child()
    _main_stack: MainStack = Gtk.Template.Child()
    _toast_overlay: Adw.ToastOverlay = Gtk.Template.Child()

    def __init__(self) -> None:
        app.window = self

        Adw.ApplicationWindow.__init__(self)
        EventHelper.__init__(self)

        self.set_application(app.app)
        self.set_title(GLib.get_application_name())
        self.set_default_icon_name("gajim")

        self._startup_finished: bool = False
        self._chat_list_visible_before_focus: bool = True

        self._restore_search_on_expand = False
        self._restore_roster_on_expand = False
        self._restore_chat_list_on_expand = False
        self._chat_list_was_collapsed = False

        self._emoji_chooser: EmojiChooser | None = None
        self._about_dialog = None
        self._chat_page = self._main_stack.get_chat_page()

        self._prepare_window()

    def init(self) -> None:
        """Init is called at a later point, so that the (empty) window can be
        displayed early, indicating a progressing Gajim startup.
        """
        self._about_dialog = AboutDialog()

        self._add_actions()
        self._add_settings_actions()
        self._add_stateful_actions()
        self._connect_actions()

        self._app_side_bar.set_chat_page(self._chat_page)

        self._split_view.connect("notify::collapsed", self._on_layout_changed)
        self._split_view.connect(
            "notify::show-sidebar", self._on_sidebar_visibility_changed
        )
        self._main_stack.connect(
            "notify::chat-list-collapsed", self._on_chat_list_collapsed_changed
        )
        self._main_stack.connect("notify::visible-child-name", self._on_layout_changed)
        self.get_action("chat-list-visible").connect(
            "notify::state", self._on_layout_changed
        )
        self._update_navigation()

        self.connect("notify::is-active", self._on_window_active)
        self.connect("close-request", self._on_close_request)

        controller = Gtk.EventControllerMotion(
            propagation_phase=Gtk.PropagationPhase.BUBBLE
        )
        controller.connect("motion", self._on_window_motion_notify)
        self.add_controller(controller)

        self.register_events(
            [
                ("message-received", ged.GUI1, self._on_message_received),
                ("read-state-sync", ged.GUI1, self._on_read_state_sync),
                ("call-started", ged.GUI1, self._on_call_started),
                ("jingle-request-received", ged.GUI1, self._on_jingle_request),
                ("file-request-received", ged.GUI1, self._on_file_request),
                ("account-enabled", ged.GUI1, self._on_account_enabled),
                ("account-disabled", ged.GUI1, self._on_account_disabled),
                ("roster-item-exchange", ged.GUI1, self._on_roster_item_exchange),
                ("plain-connection", ged.GUI1, self._on_plain_connection),
                ("password-required", ged.GUI1, self._on_password_required),
                ("http-auth", ged.GUI1, self._on_http_auth),
                ("muc-added", ged.GUI1, self._on_muc_added),
                ("message-sent", ged.GUI1, self._on_message_sent),
                ("signed-in", ged.GUI1, self._on_signed_in),
            ]
        )

        self._check_for_account()
        self._load_chats()
        self._load_unread_counts()

        app.ged.raise_event(events.RegisterActions())

        chat_list_stack = self._chat_page.get_chat_list_stack()
        app.app.systray.connect_unread_widget(chat_list_stack, "unread-count-changed")
        chat_list_stack.connect("chat-selected", self._on_chat_selected)

        for client in app.get_clients():
            client.connect_signal("state-changed", self._on_client_state_changed)
            client.connect_signal(
                "resume-successful", self._on_client_resume_successful
            )

        manager = app.app.get_shortcut_manager()
        manager.install_shortcuts(self, "main-win")

    @property
    def about_dialog(self) -> AboutDialog:
        assert self._about_dialog is not None
        return self._about_dialog

    def get_action(self, name: str) -> Gio.SimpleAction:
        action = self.lookup_action(name)
        assert isinstance(action, Gio.SimpleAction)
        return action

    def set_action_state(self, name: str, value: bool) -> None:
        action = self.lookup_action(name)
        assert action is not None
        action.change_state(GLib.Variant("b", value))

    def _on_layout_changed(self, *args: Any) -> None:
        self._update_navigation()

    def _on_sidebar_visibility_changed(self, *args: Any) -> None:
        self._update_sidebar_button()

    def _get_chat_list_visible(self) -> bool:
        if self._get_focus_mode():
            # The action is forced to False, the layout's state is kept aside
            return self._chat_list_visible_before_focus

        state = self.get_action("chat-list-visible").get_state()
        return True if state is None else state.get_boolean()

    def _set_chat_list_visible(self, visible: bool) -> None:
        # In Focus Mode the chat list stays hidden, no matter what the layout
        # asks for. Remember the requested state, so that it can be restored
        # once Focus Mode ends.
        if self._get_focus_mode():
            self._chat_list_visible_before_focus = visible
            return

        self.set_action_state("chat-list-visible", visible)

    def _on_chat_list_collapsed_changed(self, *args: Any) -> None:
        collapsed = self._main_stack.get_chat_list_collapsed()
        if collapsed == self._chat_list_was_collapsed:
            return
        self._chat_list_was_collapsed = collapsed

        if collapsed:
            self._restore_search_on_expand = self._chat_page.is_search_active()
            self._restore_roster_on_expand = not app.settings.get(
                "hide_groupchat_occupants_list"
            )
            self._chat_page.hide_search()
            app.settings.set("hide_groupchat_occupants_list", True)
            chat_stack = self._chat_page.get_chat_stack()

            # The chat stack wins over the chat list. Whatever it shows (a chat,
            # a function page, an activity) stays visible, only the placeholder
            # gives way to the list.
            if chat_stack.get_visible_child_name() != "empty":
                self._restore_chat_list_on_expand = self._get_chat_list_visible()
                self._set_chat_list_visible(False)
        else:
            if self._restore_chat_list_on_expand:
                self._set_chat_list_visible(True)
            if self._restore_roster_on_expand:
                app.settings.set("hide_groupchat_occupants_list", False)
            if self._restore_search_on_expand:
                self._chat_page.show_search()

            self._restore_chat_list_on_expand = False
            self._restore_roster_on_expand = False
            self._restore_search_on_expand = False

        self._update_navigation()

    def _update_navigation(self) -> None:
        """Adapt the layout to the current window width.

        Two Adw.Breakpoints (declared in main.ui) drive the state:
        below 800sp ``_main_stack.chat-list-collapsed`` becomes True, below
        560sp ``_split_view.collapsed`` becomes True as well. In these narrow
        bands the chat list and the conversation are shown one at a time, and
        the header bar provides a sidebar toggle.

        Focus Mode overrides all of this and hides the sidebar rail entirely.
        """
        focus_mode = self._get_focus_mode()
        rail_collapsed = self._split_view.get_collapsed()
        chat_list_collapsed = self._main_stack.get_chat_list_collapsed()
        collapsed = rail_collapsed or chat_list_collapsed

        state = self.get_action("chat-list-visible").get_state()
        showing_list = True if state is None else state.get_boolean()

        page = self._main_stack.get_visible_child_name()

        self._chat_page.set_conversation_visible(
            not chat_list_collapsed or not showing_list
        )

        self._chat_page.set_narrow(chat_list_collapsed)

        if focus_mode:
            self._split_view.set_show_sidebar(False)
        elif not rail_collapsed:
            self._split_view.set_show_sidebar(True)
        elif page == "chats":
            self._split_view.set_show_sidebar(showing_list)
        elif page == "account":
            self._split_view.set_show_sidebar(False)
        else:
            self._split_view.set_show_sidebar(True)

        self._back_button.set_visible(not focus_mode and rail_collapsed)
        self._update_sidebar_button()

        # Keep the back button reachable even if the user hid the header bar.
        if sys.platform not in ("win32", "darwin"):
            if collapsed:
                self._header_bar.set_visible(True)
            else:
                self._header_bar.set_visible(app.settings.get("show_header_bar"))

    def _update_sidebar_button(self) -> None:
        if self._split_view.get_show_sidebar():
            icon_name = "lucide-panel-left-close-symbolic"
            tooltip = _("Hide Sidebar")
        else:
            icon_name = "lucide-panel-left-open-symbolic"
            tooltip = _("Show Sidebar")
        self._back_button.set_icon_name(icon_name)
        self._back_button.set_tooltip_text(tooltip)

    @Gtk.Template.Callback()
    def _on_back_clicked(self, _button: Gtk.Button) -> None:
        show_sidebar = self._split_view.get_show_sidebar()
        self._split_view.set_show_sidebar(not show_sidebar)

    def get_chat_stack(self) -> ChatStack:
        return self._chat_page.get_chat_stack()

    def get_activity_list(self) -> ActivityListView:
        return self._chat_page.get_activity_list()

    def get_emoji_chooser(self) -> EmojiChooser:
        if self._emoji_chooser is None:
            self._emoji_chooser = EmojiChooser()

        parent = cast(Gtk.MenuButton | None, self._emoji_chooser.get_parent())
        if parent is not None:
            parent.set_popover(None)

        return self._emoji_chooser

    def show_toast(self, toast: Adw.Toast) -> None:
        self._toast_overlay.add_toast(toast)

    def show(self) -> None:
        self.present()

    def set_skip_taskbar_hint(self, value: bool) -> None:
        if not app.is_display(Display.X11):
            return
        toplevel = cast(GdkX11.X11Surface, self.get_surface())
        toplevel.set_skip_taskbar_hint(value)

    def set_urgency_hint(self, value: bool) -> None:
        if not app.settings.get("use_urgency_hint"):
            return

        if app.is_display(Display.X11):
            toplevel = cast(GdkX11.X11Surface, self.get_surface())
            toplevel.set_urgency_hint(value)
        elif app.is_display(Display.WIN32):
            toplevel = cast(GdkWin32.Win32Surface, self.get_surface())
            toplevel.set_urgency_hint(value)

    def mark_workspace_as_read(self, workspace: str) -> None:
        chat_list_stack = self._chat_page.get_chat_list_stack()
        chat_list = chat_list_stack.get_chatlist(workspace)
        open_chats = chat_list.get_open_chats()
        for chat in open_chats:
            self.mark_as_read(chat["account"], chat["jid"])

    def _mark_workspace_as_read(
        self, _action: Gio.SimpleAction, param: GLib.Variant
    ) -> None:
        workspace_id = param.get_string() or None
        if workspace_id is not None:
            self.mark_workspace_as_read(workspace_id)

    def _prepare_window(self) -> None:
        if sys.platform == "win32":
            self._header_bar.set_decoration_layout("icon:minimize,maximize,close")

        window_width = app.settings.get("mainwin_width")
        window_height = app.settings.get("mainwin_height")
        resize_window(self, window_width, window_height)

        self.set_visible(True)

        if app.is_display(Display.X11):
            self.set_skip_taskbar_hint(not app.settings.get("show_in_taskbar"))

        show_main_window = app.settings.get("show_main_window_on_startup")
        if show_main_window == "never":
            self.hide_window()

        elif show_main_window == "last_state" and not app.settings.get(
            "is_window_visible"
        ):
            self.hide_window()

    def _on_account_enabled(self, event: events.AccountEnabled) -> None:
        client = app.get_client(event.account)
        client.connect_signal("state-changed", self._on_client_state_changed)
        client.connect_signal("resume-successful", self._on_client_resume_successful)

    def _on_account_disabled(self, event: events.AccountDisabled) -> None:
        workspace_id = self._app_side_bar.get_first_workspace()
        self.activate_workspace(workspace_id)
        self._main_stack.remove_account_page(event.account)
        self._main_stack.remove_chats_for_account(event.account)

    def _on_client_state_changed(
        self, client: Client, _signal_name: str, state: SimpleClientState
    ) -> None:
        app.app.set_account_actions_state(client.account, state.is_connected)
        app.app.update_app_actions_state()

    def _on_client_resume_successful(self, client: Client, _signal_name: str) -> None:
        app.app.update_feature_actions_state(client.account)

    @staticmethod
    def _on_roster_item_exchange(event: events.RosterItemExchangeEvent) -> None:
        open_window(
            "RosterItemExchange",
            account=event.client.account,
            action=event.action,
            exchange_list=event.exchange_items_list,
            jid_from=event.jid,
        )

    @staticmethod
    def _on_plain_connection(event: events.PlainConnection) -> None:
        def _on_response(response_id: str) -> None:
            if response_id == "connect":
                event.connect()
            else:
                event.abort()

        AlertDialog(
            _("Insecure Connection"),
            _(
                "You are about to connect to the account %(account)s "
                "(%(server)s) using an insecure connection method. This means "
                "conversations will not be encrypted. Connecting PLAIN is "
                "strongly discouraged."
            )
            % {
                "account": event.account,
                "server": app.get_hostname_from_account(
                    event.account, prefer_custom=True
                ),
            },
            responses=[
                CancelDialogResponse(label=_("_Abort")),
                DialogResponse(
                    "connect", _("_Connect Anyway"), appearance="destructive"
                ),
            ],
            callback=_on_response,
        )

    @staticmethod
    def _on_password_required(event: events.PasswordRequired) -> None:
        open_window("PasswordDialog", account=event.client.account, event=event)

    @staticmethod
    def _on_http_auth(event: events.HttpAuth) -> None:
        def _on_response(response_id: str) -> None:
            event.client.get_module("HTTPAuth").build_http_auth_answer(
                event.stanza, "yes" if response_id == "accept" else "no"
            )

        account = event.client.account
        message = _("HTTP (%(method)s) Authorization for %(url)s (ID: %(id)s)") % {
            "method": event.data.method,
            "url": event.data.url,
            "id": event.data.id,
        }
        sec_msg = _("Do you accept this request?")
        if app.get_number_of_connected_accounts() > 1:
            sec_msg = _("Do you accept this request (account: %s)?") % account
        if event.data.body:
            sec_msg = event.data.body + "\n" + sec_msg
        message = message + "\n" + sec_msg

        AlertDialog(
            _("HTTP Authorization Request"),
            message,
            responses=[
                CancelDialogResponse(label=_("_No")),
                DialogResponse("accept", _("_Accept")),
            ],
            callback=_on_response,
        )

    def _on_muc_added(self, event: events.MucAdded) -> None:
        if self.chat_exists(event.account, event.jid):
            return

        self.add_group_chat(event.account, event.jid, select=event.select_chat)

    def _on_message_sent(self, event: events.MessageSent) -> None:
        if not event.play_sound:
            return

        enabled = app.settings.get_soundevent_settings("message_sent")["enabled"]
        if enabled:
            if isinstance(event.jid, list) and len(event.jid) > 1:
                return
            sound.play("message_sent", event.account)

    def _on_signed_in(self, event: events.SignedIn) -> None:
        if app.settings.get("ask_online_status"):
            self.show_account_page(event.account)

    def _add_actions(self) -> None:
        for action, variant_type, enabled in MAIN_WIN_ACTIONS:
            if variant_type is not None:
                variant_type = GLib.VariantType(variant_type)
            act = Gio.SimpleAction.new(action, variant_type)
            act.set_enabled(enabled)
            self.add_action(act)

    def _add_settings_actions(self) -> None:
        if sys.platform in ("win32", "darwin"):
            return
        action = app.settings.create_action("show_header_bar")
        action.simple_bind_property(self._header_bar, "visible")
        self.add_action(action)

    def _add_stateful_actions(self) -> None:
        action = Gio.SimpleAction.new_stateful(
            "set-encryption", GLib.VariantType("s"), GLib.Variant("s", "disabled")
        )

        self.add_action(action)

        action = Gio.SimpleAction.new_stateful(
            "chat-list-visible", None, GLib.Variant("b", True)
        )

        self.add_action(action)

        action = Gio.SimpleAction.new_stateful(
            "focus-mode", None, GLib.Variant("b", False)
        )
        action.connect("notify::state", self._on_focus_mode_changed)
        self.add_action(action)

    def _get_focus_mode(self) -> bool:
        state = self.get_action("focus-mode").get_state()
        return False if state is None else state.get_boolean()

    def _on_focus_mode_changed(self, action: Gio.SimpleAction, _param: Any) -> None:
        state = action.get_state()
        assert state is not None
        if state.get_boolean():
            chat_list_state = self.get_action("chat-list-visible").get_state()
            assert chat_list_state is not None
            self._chat_list_visible_before_focus = chat_list_state.get_boolean()
            self.set_action_state("chat-list-visible", False)
        else:
            self.set_action_state(
                "chat-list-visible", self._chat_list_visible_before_focus
            )

        self._update_navigation()

    def _connect_actions(self) -> None:
        actions = [
            ("change-nickname", self._on_action),
            ("change-subject", self._on_action),
            ("escape", self._on_action),
            ("close-chat", self._on_action),
            ("restore-chat", self._on_action),
            ("switch-next-chat", self._on_action),
            ("switch-prev-chat", self._on_action),
            ("switch-next-unread-chat", self._on_action),
            ("switch-prev-unread-chat", self._on_action),
            ("switch-chat-1", self._on_action),
            ("switch-chat-2", self._on_action),
            ("switch-chat-3", self._on_action),
            ("switch-chat-4", self._on_action),
            ("switch-chat-5", self._on_action),
            ("switch-chat-6", self._on_action),
            ("switch-chat-7", self._on_action),
            ("switch-chat-8", self._on_action),
            ("switch-chat-9", self._on_action),
            ("switch-workspace-1", self._on_action),
            ("switch-workspace-2", self._on_action),
            ("switch-workspace-3", self._on_action),
            ("switch-workspace-4", self._on_action),
            ("switch-workspace-5", self._on_action),
            ("switch-workspace-6", self._on_action),
            ("switch-workspace-7", self._on_action),
            ("switch-workspace-8", self._on_action),
            ("switch-workspace-9", self._on_action),
            ("increase-app-font-size", self._on_app_font_size_action),
            ("decrease-app-font-size", self._on_app_font_size_action),
            ("reset-app-font-size", self._on_app_font_size_action),
            ("copy-message", self._on_copy_message),
            ("retract-message", self._on_retract_message),
            ("moderate-message", self._on_moderate_message),
            ("moderate-all-messages", self._on_moderate_all_messages),
            ("delete-message-locally", self._on_delete_message_locally),
            ("add-workspace", self._add_workspace),
            ("edit-workspace", self._edit_workspace),
            ("remove-workspace", self._remove_workspace),
            ("activate-workspace", self._activate_workspace),
            ("mark-workspace-as-read", self._mark_workspace_as_read),
            ("add-chat", self._add_chat),
            ("add-group-chat", self._add_group_chat),
            ("chat-contact-info", self._on_chat_contact_info),
            ("muc-user-block", self._on_muc_user_block),
            ("muc-user-unblock", self._on_muc_user_unblock),
        ]

        for action, func in actions:
            act = self.get_action(action)
            act.connect("activate", func)

        app_actions = [
            ("handle-uri", self._on_handle_uri),
        ]

        for action, func in app_actions:
            act = app.app.lookup_action(action)
            assert act is not None
            act.connect("activate", func)

    def _on_action(
        self, action: Gio.SimpleAction, _param: GLib.Variant | None
    ) -> int | None:
        action_name = action.get_name()
        log.info("Activate action: %s", action_name)

        if action_name == "escape" and self._chat_page.hide_search():
            return None

        chat_stack = self._chat_page.get_chat_stack()
        if action_name == "escape" and chat_stack.process_escape():
            return None

        control = self.get_control()
        if control.has_active_chat():
            contact = control.get_contact()
            assert contact is not None

            if action_name == "change-nickname":
                app.window.activate_action("win.muc-change-nickname", None)
                return None

            if action_name == "change-subject":
                open_window("GroupchatDetails", contact=contact, page="manage")
                return None

            if action_name == "escape":
                if app.settings.get("escape_key_closes"):
                    self._chat_page.remove_chat(contact.account, contact.jid)
                else:
                    workspace_id = self.get_active_workspace()
                    if workspace_id is not None:
                        app.window.activate_workspace(workspace_id)

                return None

            elif action_name == "close-chat":
                self._chat_page.remove_chat(contact.account, contact.jid)
                return None

        if action_name == "escape" and app.settings.get("escape_key_closes"):
            self.close()

        elif action_name == "close-chat":
            self.close()

        elif action_name == "restore-chat":
            self._chat_page.restore_chat()

        elif action_name == "switch-next-chat":
            self.select_next_chat(Direction.NEXT)

        elif action_name == "switch-prev-chat":
            self.select_next_chat(Direction.PREV)

        elif action_name == "switch-next-unread-chat":
            self.select_next_chat(Direction.NEXT, unread_first=True)

        elif action_name == "switch-prev-unread-chat":
            self.select_next_chat(Direction.PREV, unread_first=True)

        elif action_name.startswith("switch-chat-"):
            number = int(action_name[-1]) - 1
            self.select_chat_number(number)

        elif action_name.startswith("switch-workspace-"):
            number = int(action_name[-1]) - 1
            self._app_side_bar.activate_workspace_number(number)

        return None

    def _on_app_font_size_action(
        self, action: Gio.SimpleAction, _param: GLib.Variant
    ) -> None:
        action_name = action.get_name()
        if action_name == "reset-app-font-size":
            app.settings.set_app_setting("app_font_size", None)
            app.css_config.apply_app_font_size()
            return

        app_font_size = app.settings.get("app_font_size")
        new_app_font_size = app_font_size
        if action_name == "increase-app-font-size":
            new_app_font_size = app_font_size + 0.10
        elif action_name == "decrease-app-font-size":
            new_app_font_size = app_font_size - 0.10

        # Clamp font size
        new_app_font_size = max(min(3.0, new_app_font_size), 0.5)

        if new_app_font_size == app_font_size:
            return

        app.settings.set_app_setting("app_font_size", new_app_font_size)
        app.css_config.apply_app_font_size()

    def _on_copy_message(self, _action: Gio.SimpleAction, param: GLib.Variant) -> None:
        self.get_clipboard().set(param.get_string())

    @actionmethod
    def _on_chat_contact_info(
        self, _action: Gio.SimpleAction, params: AccountJidParam
    ) -> None:
        client = app.get_client(params.account)
        contact = client.get_module("Contacts").get_contact(params.jid)
        if isinstance(contact, GroupchatContact):
            open_window("GroupchatDetails", contact=contact)
        else:
            open_window("ContactInfo", account=contact.account, contact=contact)

    @actionmethod
    def _on_muc_user_block(
        self, _action: Gio.SimpleAction, params: OccupantParam
    ) -> None:
        def _on_response() -> None:
            client = app.get_client(params.account)
            client.get_module("MucBlocking").set_block_occupants(
                params.jid, [params.occupant_id], True
            )

            # Close PM chats
            contact = client.get_module("Contacts").get_contact(
                params.jid, groupchat=True
            )
            assert isinstance(contact, GroupchatContact)
            participant = contact.get_occupant(params.occupant_id)
            if participant is None:
                return

            self._chat_page.remove_chat(params.account, participant.jid)
            app.app.avatar_storage.remove_avatar(participant)
            client.get_module("VCardAvatars").invalidate_cache(participant.jid)
            participant.update_avatar()

        ConfirmationAlertDialog(
            _("Block Participant?"),
            _("Do you want to block %(name)s?") % {"name": params.resource},
            confirm_label=_("_Block"),
            callback=_on_response,
        )

    @actionmethod
    def _on_muc_user_unblock(
        self, _action: Gio.SimpleAction, params: OccupantParam
    ) -> None:
        def _on_response() -> None:
            client = app.get_client(params.account)
            client.get_module("MucBlocking").set_block_occupants(
                params.jid, [params.occupant_id], False
            )

        ConfirmationAlertDialog(
            _("Unblock Participant?"),
            _("Do you want to unblock %(name)s?") % {"name": params.resource},
            confirm_label=_("_Unblock"),
            callback=_on_response,
        )

    @actionmethod
    def _on_retract_message(
        self, _action: Gio.SimpleAction, params: RetractMessageParam
    ) -> None:
        def _on_response() -> None:
            client = app.get_client(params.account)
            contact = client.get_module("Contacts").get_contact(params.jid)
            assert not isinstance(contact, ResourceContact)
            client.get_module("Retraction").send_retraction(contact, params.retract_ids)

        ConfirmationAlertDialog(
            _("Retract Message?"),
            _(
                "Do you want to retract this message?\n"
                "Please note that retracting a message does not guarantee that your "
                "provider or your contact’s device will remove it."
            ),
            confirm_label=_("_Retract"),
            appearance="destructive",
            callback=_on_response,
        )

    @actionmethod
    def _on_moderate_message(
        self, _action: Gio.SimpleAction, params: ModerateMessageParam
    ) -> None:
        def _on_response(reason: str) -> None:
            client = app.get_client(params.account)
            client.get_module("MUC").moderate_messages(
                params.namespace, params.jid, params.stanza_ids, reason or None
            )

        ConfirmationAlertDialog(
            _("Moderate Message?"),
            _("Why do you want to moderate this message?"),
            confirm_label=_("_Moderate"),
            appearance="destructive",
            extra_widget=DialogEntry(text=_("Spam")),
            callback=_on_response,
        )

    @actionmethod
    def _on_moderate_all_messages(
        self, _action: Gio.SimpleAction, params: ModerateAllMessagesParam
    ) -> None:
        stanza_ids = app.storage.archive.get_message_stanza_ids_from_occupant(
            params.account, params.jid, params.occupant_id
        )
        if stanza_ids is None:
            InformationAlertDialog(
                _("No Messages Found"),
                _("Could not find any messages for this participant."),
            )
            return

        def _on_response(reason: str) -> None:
            client = app.get_client(params.account)
            groupchat_contact = client.get_module("Contacts").get_contact(params.jid)
            assert isinstance(groupchat_contact, GroupchatContact)
            if not groupchat_contact.is_joined:
                InformationAlertDialog(
                    _("Not Joined"), _("You are currently not joined this group chat")
                )
                return

            muc_module = client.get_module("MUC")
            muc_module.moderate_messages(
                params.namespace, params.jid, stanza_ids, reason or None
            )

        ConfirmationAlertDialog(
            _("Moderate Messages?"),
            _(
                "%(count)s messages from %(participant)s will be moderated.\n"
                "Why do you want to moderate these messages?"
            )
            % {"count": len(stanza_ids), "participant": params.nickname},
            confirm_label=_("_Moderate"),
            appearance="destructive",
            extra_widget=DialogEntry(text=_("Spam")),
            callback=_on_response,
        )

    @actionmethod
    def _on_delete_message_locally(
        self, _action: Gio.SimpleAction, params: DeleteMessageParam
    ) -> None:
        def _on_response() -> None:
            app.storage.archive.delete_message(params.pk)
            app.ged.raise_event(
                events.MessageDeleted(
                    account=params.account, jid=params.jid, pk=params.pk
                )
            )

        ConfirmationAlertDialog(
            _("Delete Message Locally?"),
            _("This message will be deleted from your local chat history"),
            confirm_label=_("_Delete"),
            appearance="destructive",
            callback=_on_response,
        )

    def _on_handle_uri(self, _action: Gio.SimpleAction, param: GLib.Variant) -> None:
        uris = param.unpack()
        log.debug("Try to handle uris: %s", uris)
        if not uris:
            return

        self.present()

        uri = parse_uri(uris[0])
        match uri:
            case XmppIri():
                self.open_xmpp_iri(uri)
            case InvalidUri():
                log.error("Failed to handle uri: %s", uri.error)
            case _:
                log.warning("Can only handle xmpp iris: %s", uris[0])

    def open_xmpp_iri(self, xmpp_iri: XmppIri) -> None:
        accounts = app.settings.get_active_accounts()
        if not accounts:
            log.warning("No accounts active, unable to handle uri")
            return

        jid_str = str(xmpp_iri.jid)

        match xmpp_iri.action:
            case "join":
                if len(accounts) == 1:
                    self.activate_action(
                        "app.open-chat", GLib.Variant("as", [accounts[0], jid_str])
                    )
                else:
                    self.activate_action(
                        "app.start-chat", GLib.Variant("as", [jid_str, ""])
                    )

            case "roster":
                self.activate_action("app.add-contact", GLib.Variant("s", jid_str))

            case "message" | "":
                body = xmpp_iri.params.get("body")
                app.window.start_chat_from_jid(accounts[0], jid_str, body or None)

            case _:
                log.warning("No handler for action: %s", xmpp_iri)

    def _on_window_motion_notify(
        self,
        _widget: Gtk.EventControllerMotion,
        _x: float,
        _y: float,
    ) -> None:
        control = self.get_control()
        if not control.has_active_chat():
            return

        if self.is_active():
            contact = control.get_contact()
            assert contact is not None
            client = app.get_client(contact.account)
            chat_stack = self._chat_page.get_chat_stack()
            client.get_module("Chatstate").set_mouse_activity(
                contact, chat_stack.get_message_input().has_text
            )

    def _on_chat_selected(self, *args: Any) -> None:
        self._app_side_bar.select_chat()

        if self._main_stack.get_chat_list_collapsed():
            self._set_chat_list_visible(False)

    def _on_close_request(self, _widget: Gtk.ApplicationWindow) -> int:
        if app.settings.get("confirm_on_window_delete"):
            open_window("QuitDialog")
            return Gdk.EVENT_STOP

        action = app.settings.get("action_on_close")
        if action == "hide":
            self.hide_window()
        elif action == "minimize":
            self.minimize()
        else:
            app.app.start_shutdown()

        return Gdk.EVENT_STOP

    def show_window(self) -> None:
        app.settings.set("is_window_visible", True)
        self.set_visible(True)

    def hide_window(self) -> None:
        app.settings.set("is_window_visible", False)
        self.set_visible(False)

    def _set_startup_finished(self) -> None:
        self._startup_finished = True
        self._chat_page.set_startup_finished()

    def _load_unread_counts(self) -> None:
        chats = app.storage.cache.get_unread()
        chat_list_stack = self._chat_page.get_chat_list_stack()

        for chat in chats:
            chat_list_stack.set_chat_unread_count(chat.account, chat.jid, chat.count)

    def show_account_page(self, account: str) -> None:
        self._app_side_bar.show_account_page()
        self._main_stack.show_account(account)

    def get_active_workspace(self) -> str | None:
        return self._app_side_bar.get_active_workspace()

    def is_chat_active(self, account: str, jid: JID) -> bool:
        if not self.is_active():
            return False
        if self._main_stack.get_visible_page_name() != "chats":
            return False
        if not self._chat_page.is_chat_visible():
            return False
        return self._chat_page.is_chat_selected(account, jid)

    def highlight_dnd_targets(self, dragged_object: Any, highlight: bool) -> None:
        if isinstance(dragged_object, ChatListRow | PreviewWidget):
            chat_list_stack = self._chat_page.get_chat_list_stack()
            workspace = self.get_active_workspace()
            if workspace is None:
                return

            chat_list = chat_list_stack.get_chatlist(workspace)
            if isinstance(dragged_object, ChatListRow) and dragged_object.is_pinned:
                for row in chat_list.get_chat_list_rows():
                    if not row.is_pinned:
                        continue

                    if highlight:
                        row.add_css_class("dnd-target-chatlist")
                    else:
                        row.remove_css_class("dnd-target-chatlist")

            if isinstance(dragged_object, PreviewWidget):
                for row in chat_list.get_chat_list_rows():
                    if highlight:
                        row.add_css_class("dnd-target-chatlist")
                    else:
                        row.remove_css_class("dnd-target-chatlist")

        self._app_side_bar.highlight_dnd_targets(dragged_object, highlight)

    def _add_workspace(self, _action: Gio.SimpleAction, param: GLib.Variant) -> None:
        workspace_id = param.get_string()
        if workspace_id:
            self.add_workspace(workspace_id)

    def add_workspace(
        self, workspace_id: str | None = None, switch: bool = True
    ) -> str:
        if workspace_id is None:
            workspace_id = app.settings.add_workspace(_("My Workspace"))

        self._app_side_bar.add_workspace(workspace_id)
        self._chat_page.add_chat_list(workspace_id)

        if self._startup_finished and switch:
            self.activate_workspace(workspace_id)
            self._app_side_bar.store_workspace_order()

        return workspace_id

    def _edit_workspace(self, _action: Gio.SimpleAction, param: GLib.Variant) -> None:
        workspace_id = param.get_string() or None
        if workspace_id is None:
            workspace_id = self.get_active_workspace()
        open_window("WorkspaceDialog", workspace_id=workspace_id)

    def _remove_workspace(self, _action: Gio.SimpleAction, param: GLib.Variant) -> None:
        workspace_id = param.get_string() or None
        if workspace_id is None:
            workspace_id = self.get_active_workspace()

        if workspace_id is not None:
            self.remove_workspace(workspace_id)

    def remove_workspace(self, workspace_id: str) -> None:
        if app.settings.get_workspace_count() == 1:
            log.warning("Tried to remove the only workspace")
            return

        was_active = self.get_active_workspace() == workspace_id
        chat_list = self.get_chat_list(workspace_id)
        open_chats = chat_list.get_open_chats()

        def _continue_removing_workspace():
            new_workspace_id = self._app_side_bar.get_other_workspace(workspace_id)
            if new_workspace_id is None:
                log.warning("No other workspaces found")
                return

            for open_chat in open_chats:
                params = ChatListEntryParam(
                    workspace_id=new_workspace_id,
                    source_workspace_id=workspace_id,
                    account=open_chat["account"],
                    jid=open_chat["jid"],
                )
                self.activate_action("move-chat-to-workspace", params.to_variant())

            if was_active:
                self.activate_workspace(new_workspace_id)

            self._app_side_bar.remove_workspace(workspace_id)
            self._chat_page.remove_chat_list(workspace_id)
            app.settings.remove_workspace(workspace_id)

        if open_chats:
            ConfirmationAlertDialog(
                _("Remove Workspace?"),
                _(
                    "This workspace contains chats. All chats will be moved to "
                    "the next workspace. Remove anyway?"
                ),
                confirm_label=_("_Remove"),
                appearance="destructive",
                callback=_continue_removing_workspace,
            )
            return

        # No chats in chat list, it is save to remove this workspace
        _continue_removing_workspace()

    def _activate_workspace(
        self, _action: Gio.SimpleAction, param: GLib.Variant
    ) -> None:
        workspace_id = param.get_string()
        if workspace_id:
            self.activate_workspace(workspace_id)

    def activate_workspace(self, workspace_id: str) -> None:
        self._app_side_bar.activate_workspace(workspace_id)
        self._main_stack.show_chats(workspace_id)

        self.set_action_state("focus-mode", False)
        self.set_action_state("chat-list-visible", True)

    def update_workspace(self, workspace_id: str) -> None:
        self._chat_page.update_workspace(workspace_id)
        self._app_side_bar.update_workspace(workspace_id)

    def get_chat_list(self, workspace_id: str) -> ChatList:
        chat_list_stack = self._chat_page.get_chat_list_stack()
        return chat_list_stack.get_chatlist(workspace_id)

    def _get_suitable_workspace(
        self, account: str, jid: JID, private_chat: bool = False
    ) -> str:
        if private_chat:
            # Try to add private chat to the same workspace the MUC resides in
            chat_list_stack = self._chat_page.get_chat_list_stack()
            chat_list = chat_list_stack.find_chat(account, jid.new_as_bare())
            if chat_list is not None:
                return chat_list.workspace_id

        default = app.settings.get_account_setting(account, "default_workspace")
        workspaces = app.settings.get_workspaces()
        if default in workspaces:
            return default

        workspace_id = self.get_active_workspace()
        if workspace_id is not None:
            return workspace_id
        return self._app_side_bar.get_first_workspace()

    def _add_group_chat(self, _action: Gio.SimpleAction, param: GLib.Variant) -> None:
        account, jid, select = param.unpack()
        self.add_group_chat(account, JID.from_string(jid), select)

    def add_group_chat(self, account: str, jid: JID, select: bool = False) -> None:
        workspace_id = self._get_suitable_workspace(account, jid)
        self._chat_page.add_chat_for_workspace(
            workspace_id, account, jid, "groupchat", select=select
        )

    def _add_chat(self, _action: Gio.SimpleAction, param: GLib.Variant) -> None:
        params = AddChatActionParams.from_variant(param)
        self.add_chat(params.account, params.jid, params.type, params.select)

    def add_chat(
        self,
        account: str,
        jid: JID,
        type_: Literal["chat", "groupchat", "pm"],
        select: bool = False,
        message: str | None = None,
    ) -> None:
        workspace_id = self._get_suitable_workspace(account, jid)
        self._chat_page.add_chat_for_workspace(
            workspace_id, account, jid, type_, select=select, message=message
        )

    def add_private_chat(self, account: str, jid: JID, select: bool = False) -> None:
        workspace_id = self._get_suitable_workspace(account, jid, private_chat=True)
        self._chat_page.add_chat_for_workspace(
            workspace_id, account, jid, "pm", select=select
        )

    def clear_chat_list_row(self, account: str, jid: JID) -> None:
        chat_list_stack = self._chat_page.get_chat_list_stack()
        chat_list = chat_list_stack.find_chat(account, jid)
        if chat_list is not None:
            chat_list.clear_chat_list_row(account, jid)

    def select_chat(self, account: str, jid: JID) -> None:
        self._app_side_bar.select_chat()
        self._main_stack.show_chat_page()
        self._chat_page.select_chat(account, jid)

    def scroll_to_message(self, account: str, message: Message) -> None:
        control = self._chat_page.get_control()
        control.set_prepare_for_scroll()

        if not self.is_chat_active(account, message.remote.jid):
            message_type = MessageType(message.type).to_str()
            app.window.add_chat(account, message.remote.jid, message_type, select=True)

        if self._main_stack.get_chat_list_collapsed():
            self._chat_page.hide_search()

        control.scroll_to_message(message.pk, message.timestamp)

    def select_next_chat(
        self, direction: Direction, unread_first: bool = False
    ) -> None:
        chat_list_stack = self._chat_page.get_chat_list_stack()
        chat_list = chat_list_stack.get_current_chat_list()
        if chat_list is not None:
            chat_list.select_next_chat(direction, unread_first)

    def select_chat_number(self, number: int) -> None:
        chat_list_stack = self._chat_page.get_chat_list_stack()
        chat_list = chat_list_stack.get_current_chat_list()
        if chat_list is not None:
            chat_list.select_chat_number(number)

    def show_activity_page(self, context_id: str | None = None) -> None:
        self._app_side_bar.show_activity_page()
        self._main_stack.show_activity_page(context_id)

        self.set_action_state("focus-mode", False)
        self.set_action_state("chat-list-visible", True)

    def get_control(self) -> ChatControl:
        return self._chat_page.get_control()

    def chat_exists(self, account: str, jid: JID) -> bool:
        return self._chat_page.chat_exists(account, jid)

    def get_total_unread_count(self) -> int:
        chat_list_stack = self._chat_page.get_chat_list_stack()
        return chat_list_stack.get_total_unread_count()

    def get_chat_unread_count(
        self, account: str, jid: JID, include_silent: bool = False
    ) -> int:
        chat_list_stack = self._chat_page.get_chat_list_stack()
        count = chat_list_stack.get_chat_unread_count(account, jid, include_silent)
        return count or 0

    def mark_as_read(
        self,
        account: str,
        jid: JID,
        *,
        is_sync: bool = False,
    ) -> None:
        unread_count = self.get_chat_unread_count(account, jid, include_silent=True)

        self.set_urgency_hint(False)
        control = self.get_control()
        if control.has_active_chat():
            # Reset jump to bottom button unread counter
            control.mark_as_read()

        # Reset chat list unread counter (emits unread-count-changed)
        chat_list_stack = self._chat_page.get_chat_list_stack()
        chat_list_stack.mark_as_read(account, jid)

        # Mark reaction notifications from contact as read
        activity_list = self._chat_page.get_activity_list()
        activity_list.mark_items_as_read(Reaction, account, jid)

        if not unread_count or is_sync:
            # Read marker must be sent only once
            return

        if not app.account_is_connected(account):
            return

        client = app.get_client(account)
        contact = client.get_module("Contacts").get_contact(jid)
        assert not isinstance(contact, ResourceContact)

        if not client.get_module("MAM").is_catch_up_finished(contact):
            # Dont set read state before we have all messages
            return

        last_message = app.storage.archive.get_last_conversation_row(
            account, jid, incl_related_data=False
        )
        if last_message is None or last_message.id is None:
            return

        mds_assist_sent = client.get_module("ChatMarkers").send_displayed_marker(
            contact, last_message.id, last_message.stanza_id
        )

        if not mds_assist_sent and last_message.stanza_id is not None:
            by = contact.jid if isinstance(contact, GroupchatContact) else None
            client.get_module("MDS").set_mds(contact.jid, last_message.stanza_id, by)

    def _on_window_active(self, window: Gtk.ApplicationWindow, _param: Any) -> None:
        if not window.is_active():
            return

        self.set_urgency_hint(False)
        control = self.get_control()
        if not control.has_active_chat():
            return

        contact = control.get_contact()
        assert contact is not None

        if control.view_is_at_bottom():
            self.mark_as_read(contact.account, contact.jid)

    def get_preferred_ft_method(self, contact: types.ChatContactT) -> str | None:
        httpupload_enabled = app.window.get_action_enabled("send-file-httpupload")
        return "httpupload" if httpupload_enabled else None
        # TODO Jingle FT
        # ft_pref = app.settings.get_account_setting(
        #     contact.account,
        #     'filetransfer_preference')

        # jingle_enabled = app.window.get_action_enabled('send-file-jingle')

        # if isinstance(contact, GroupchatContact):
        #     if httpupload_enabled:
        #         return 'httpupload'
        #     return None

        # if httpupload_enabled and jingle_enabled:
        #     return ft_pref

        # if httpupload_enabled:
        #     return 'httpupload'

        # if jingle_enabled:
        #     return 'jingle'

        # return None

    def show_add_join_groupchat(
        self,
        account: str,
        jid: str,
        nickname: str | None = None,
        password: str | None = None,
    ) -> None:
        jid_ = JID.from_string(jid)
        if not self.chat_exists(account, jid_):
            client = app.get_client(account)
            client.get_module("MUC").join(jid_, nick=nickname, password=password)

        self.add_group_chat(account, jid_, select=True)

    def start_chat_from_jid(
        self, account: str, jid: str, message: str | None = None
    ) -> None:
        jid_ = JID.from_string(jid)
        if self.chat_exists(account, jid_):
            self.select_chat(account, jid_)
            if message is not None:
                message_input = self.get_chat_stack().get_message_input()
                message_input.insert_text(message)
            return

        app.app.activate_action(
            "start-chat", GLib.Variant("as", [str(jid), message or ""])
        )

    def block_contact(self, account: str, jid: JID) -> None:
        client = app.get_client(account)

        contact = client.get_module("Contacts").get_contact(jid)
        assert isinstance(contact, BareContact)
        if contact.is_blocked:
            client.get_module("Blocking").unblock([contact.jid])
            return

        # TODO: Keep "confirm_block" setting?
        def _on_response(response_id: str) -> None:
            if response_id not in ("report", "block"):
                return
            report = "spam" if response_id == "report" else None
            client.get_module("Blocking").block([contact.jid], report)
            self._chat_page.remove_chat(account, contact.jid)

        AlertDialog(
            _("Block Contact?"),
            _(
                "You will appear offline for this contact and you "
                "will not receive further messages."
            ),
            responses=[
                CancelDialogResponse(),
                DialogResponse("report", _("_Report Spam")),
                DialogResponse("block", _("_Block")),
            ],
            callback=_on_response,
        )

    def remove_contact(self, account: str, jid: JID) -> None:
        client = app.get_client(account)

        def _on_response() -> None:
            client.get_module("Roster").delete_item(jid)

        contact = client.get_module("Contacts").get_contact(jid)
        assert isinstance(
            contact, BareContact | GroupchatContact | GroupchatParticipant
        )
        sec_text = _(
            "You are about to remove %(name)s (%(jid)s) from your contact list."
        ) % {"name": contact.name, "jid": jid}

        ConfirmationAlertDialog(
            _("Remove From Contact List?"),
            sec_text,
            confirm_label=_("_Remove"),
            appearance="destructive",
            callback=_on_response,
        )

    @staticmethod
    def _check_for_account() -> None:
        accounts = app.settings.get_accounts()
        if not accounts:

            def _open_wizard():
                open_window("AccountWizard")

            GLib.idle_add(_open_wizard)

    def _load_chats(self) -> None:
        for workspace_id in app.settings.get_workspaces():
            self.add_workspace(workspace_id)
            self._chat_page.load_workspace_chats(workspace_id)

        workspace_id = self._app_side_bar.get_first_workspace()
        self.activate_workspace(workspace_id)

        self._set_startup_finished()

    def _is_history_sync(self, event: events.MessageReceived) -> bool:
        if event.mam is None:
            return False

        win = get_app_window("HistorySyncAssistant", account=event.account)
        if win is None:
            return False

        return win.get_active_query_id() == event.mam.query_id

    def _on_message_received(self, event: events.MessageReceived) -> bool | None:
        if self._is_history_sync(event):
            return ged.STOP_PROPAGATION

        if self.chat_exists(event.account, event.jid):
            return

        if event.m_type == MessageType.PM:
            if event.message.occupant is not None:
                client = app.get_client(event.account)
                jid = event.message.remote.jid.new_as_bare()
                if client.get_module("MucBlocking").is_blocked(
                    jid, event.message.occupant.id
                ):
                    log.info("PM blocked from %s", jid)
                    return ged.STOP_PROPAGATION

            self.add_private_chat(event.account, event.jid)

        else:
            self.add_chat(event.account, event.jid, "chat")

    def _on_read_state_sync(self, event: events.ReadStateSync) -> None:
        last_message = app.storage.archive.get_last_conversation_row(
            event.account, event.jid, incl_related_data=False
        )

        if last_message is None:
            return

        if event.marker_id not in (last_message.id, last_message.stanza_id):
            return

        self.mark_as_read(event.account, event.jid, is_sync=True)

    def _on_call_started(self, event: events.CallStarted) -> None:
        # Make sure there is only one window
        win = get_app_window("CallWindow")
        if win is not None:
            win.close()
        open_window(
            "CallWindow", account=event.account, resource_jid=event.resource_jid
        )

    def _on_jingle_request(self, event: events.JingleRequestReceived) -> None:
        if not self.chat_exists(event.account, event.jid):
            for item in event.contents:
                if item.media not in ("audio", "video"):
                    return
                self.add_chat(event.account, event.jid, "chat")
                break

    def _on_file_request(self, event: events.FileRequestReceivedEvent) -> None:
        if not self.chat_exists(event.account, event.jid):
            self.add_chat(event.account, event.jid, "chat")

    def start_shutdown(self) -> None:
        self.show_toast(Adw.Toast(title=_("Gajim is quitting…"), timeout=0))
        app.ged.raise_event(events.PrepareForShutdown())

        if self.is_visible():
            window_width, window_height = self.get_width(), self.get_height()
            app.settings.set("mainwin_width", window_width)
            app.settings.set("mainwin_height", window_height)

    def reload_view(self, *args: Any) -> None:
        self.get_control().get_conversation_view().reload()



# This file is part of Gajim.
#
# SPDX-License-Identifier: GPL-3.0-only

from __future__ import annotations

from typing import Any
from typing import Literal
from typing import overload
from typing import TYPE_CHECKING

import logging
from importlib import import_module

from gi.repository import Gdk
from gi.repository import Gtk
from nbxmpp import JID

from gajim.common import app

from gajim.gtk.const import WINDOW_MODULES

if TYPE_CHECKING:
    from gajim.gtk.account_wizard import AccountWizard
    from gajim.gtk.add_contact import AddContact
    from gajim.gtk.adhoc import AdHocCommands
    from gajim.gtk.advanced_config import AdvancedConfig
    from gajim.gtk.call_window import CallWindow
    from gajim.gtk.certificate_dialog import CertificateDialog
    from gajim.gtk.change_password import ChangePassword
    from gajim.gtk.contact_info import ContactInfo
    from gajim.gtk.db_migration import DBMigration
    from gajim.gtk.debug_console import DebugConsoleWindow
    from gajim.gtk.discovery import ServiceDiscoveryWindow
    from gajim.gtk.features import Features
    from gajim.gtk.groupchat_creation import CreateGroupchatWindow
    from gajim.gtk.groupchat_details import GroupchatDetails
    from gajim.gtk.groupchat_join import GroupchatJoin
    from gajim.gtk.history_export import HistoryExport
    from gajim.gtk.history_sync import HistorySyncAssistant
    from gajim.gtk.manage_proxies import ManageProxies
    from gajim.gtk.manage_sounds import ManageSounds
    from gajim.gtk.openpgp_secret import OpenPGPSecret
    from gajim.gtk.openpgp_wizard import OpenPGPWizard
    from gajim.gtk.password_dialog import PasswordDialog
    from gajim.gtk.pep_config import PEPConfig
    from gajim.gtk.preference.dialog import Preferences
    from gajim.gtk.profile import ProfileWindow
    from gajim.gtk.quit import QuitDialog
    from gajim.gtk.remove_account import RemoveAccount
    from gajim.gtk.roster_item_exchange import RosterItemExchange
    from gajim.gtk.service_registration import ServiceRegistration
    from gajim.gtk.ssl_error_dialog import SSLErrorDialog
    from gajim.gtk.start_chat import StartChatDialog
    from gajim.gtk.themes import Themes
    from gajim.gtk.workspace_dialog import WorkspaceDialog

    GajimWindowT = (
        AccountWizard
        | AddContact
        | AdHocCommands
        | AdvancedConfig
        | DBMigration
        | DebugConsoleWindow
        | CallWindow
        | CertificateDialog
        | ChangePassword
        | ContactInfo
        | CreateGroupchatWindow
        | Features
        | GroupchatDetails
        | GroupchatJoin
        | HistoryExport
        | HistorySyncAssistant
        | ManageProxies
        | ManageSounds
        | OpenPGPSecret
        | OpenPGPWizard
        | PasswordDialog
        | PEPConfig
        | Preferences
        | ProfileWindow
        | QuitDialog
        | RemoveAccount
        | RosterItemExchange
        | ServiceDiscoveryWindow
        | ServiceRegistration
        | SSLErrorDialog
        | StartChatDialog
        | Themes
        | WorkspaceDialog
    )
    GajimWindowNameT = (
        Literal["AccountWizard"]
        | Literal["AddContact"]
        | Literal["AdHocCommands"]
        | Literal["AdvancedConfig"]
        | Literal["DBMigration"]
        | Literal["DebugConsoleWindow"]
        | Literal["CallWindow"]
        | Literal["CertificateDialog"]
        | Literal["ChangePassword"]
        | Literal["ContactInfo"]
        | Literal["CreateGroupchatWindow"]
        | Literal["Features"]
        | Literal["GroupchatDetails"]
        | Literal["GroupchatJoin"]
        | Literal["HistoryExport"]
        | Literal["HistorySyncAssistant"]
        | Literal["ManageProxies"]
        | Literal["ManageSounds"]
        | Literal["OpenPGPSecret"]
        | Literal["OpenPGPWizard"]
        | Literal["PasswordDialog"]
        | Literal["PEPConfig"]
        | Literal["Preferences"]
        | Literal["ProfileWindow"]
        | Literal["QuitDialog"]
        | Literal["RemoveAccount"]
        | Literal["RosterItemExchange"]
        | Literal["ServiceDiscoveryWindow"]
        | Literal["ServiceRegistration"]
        | Literal["SSLErrorDialog"]
        | Literal["StartChatDialog"]
        | Literal["Themes"]
        | Literal["WorkspaceDialog"]
    )

log = logging.getLogger("gajim.gtk.util.window")


def get_total_screen_geometry() -> tuple[int, int]:
    total_width = 0
    total_height = 0
    display = Gdk.Display.get_default()
    assert display is not None
    monitors = display.get_monitors()
    for num in range(monitors.get_n_items()):
        monitor = monitors.get_item(num)
        assert isinstance(monitor, Gdk.Monitor)
        geometry = monitor.get_geometry()
        total_width += geometry.width
        total_height = max(total_height, geometry.height)
    log.debug("Get screen geometry: %s %s", total_width, total_height)
    return total_width, total_height


def resize_window(window: Gtk.Window, width: int, height: int) -> None:
    """
    Resize window, but also checks if huge window or negative values
    """
    screen_w, screen_h = get_total_screen_geometry()
    if not width or not height:
        return

    width = min(width, screen_w)
    height = min(height, screen_h)
    window.set_default_size(abs(width), abs(height))


def get_app_windows(account: str) -> list[Gtk.Window]:
    windows: list[Gtk.Window] = []
    for win in app.app.get_windows():
        if hasattr(win, "wrapper"):
            win = win.wrapper  # type: ignore

        if hasattr(win, "account"):  # type: ignore
            if win.account == account:  # type: ignore
                windows.append(win)  # type: ignore
    return windows


@overload
def get_app_window(
    name: Literal["AccountWizard"],
    account: str | None = None,
    jid: str | JID | None = None,
) -> AccountWizard | None: ...


@overload
def get_app_window(
    name: Literal["AddContact"],
    account: str | None = None,
    jid: str | JID | None = None,
) -> AddContact | None: ...


@overload
def get_app_window(
    name: Literal["AdHocCommands"],
    account: str | None = None,
    jid: str | JID | None = None,
) -> AdHocCommands | None: ...


@overload
def get_app_window(
    name: Literal["AdvancedConfig"],
    account: str | None = None,
    jid: str | JID | None = None,
) -> AdvancedConfig | None: ...


@overload
def get_app_window(
    name: Literal["DBMigration"],
    account: str | None = None,
    jid: str | JID | None = None,
) -> DBMigration | None: ...


@overload
def get_app_window(
    name: Literal["DebugConsoleWindow"],
    account: str | None = None,
    jid: str | JID | None = None,
) -> DebugConsoleWindow | None: ...


@overload
def get_app_window(
    name: Literal["CallWindow"],
    account: str | None = None,
    jid: str | JID | None = None,
) -> CallWindow | None: ...


@overload
def get_app_window(
    name: Literal["CertificateDialog"],
    account: str | None = None,
    jid: str | JID | None = None,
) -> CertificateDialog | None: ...


@overload
def get_app_window(
    name: Literal["ChangePassword"],
    account: str | None = None,
    jid: str | JID | None = None,
) -> ChangePassword | None: ...


@overload
def get_app_window(
    name: Literal["ContactInfo"],
    account: str | None = None,
    jid: str | JID | None = None,
) -> ContactInfo | None: ...


@overload
def get_app_window(
    name: Literal["CreateGroupchatWindow"],
    account: str | None = None,
    jid: str | JID | None = None,
) -> CreateGroupchatWindow | None: ...


@overload
def get_app_window(
    name: Literal["Features"], account: str | None = None, jid: str | JID | None = None
) -> Features | None: ...


@overload
def get_app_window(
    name: Literal["GroupchatDetails"],
    account: str | None = None,
    jid: str | JID | None = None,
) -> GroupchatDetails | None: ...


@overload
def get_app_window(
    name: Literal["GroupchatJoin"],
    account: str | None = None,
    jid: str | JID | None = None,
) -> GroupchatJoin | None: ...


@overload
def get_app_window(
    name: Literal["HistoryExport"],
    account: str | None = None,
    jid: str | JID | None = None,
) -> HistoryExport | None: ...


@overload
def get_app_window(
    name: Literal["HistorySyncAssistant"],
    account: str | None = None,
    jid: str | JID | None = None,
) -> HistorySyncAssistant | None: ...


@overload
def get_app_window(
    name: Literal["ManageProxies"],
    account: str | None = None,
    jid: str | JID | None = None,
) -> ManageProxies | None: ...


@overload
def get_app_window(
    name: Literal["ManageSounds"],
    account: str | None = None,
    jid: str | JID | None = None,
) -> ManageSounds | None: ...


@overload
def get_app_window(
    name: Literal["OpenPGPSecret"],
    account: str | None = None,
    jid: str | JID | None = None,
) -> OpenPGPWizard | None: ...


@overload
def get_app_window(
    name: Literal["OpenPGPWizard"],
    account: str | None = None,
    jid: str | JID | None = None,
) -> OpenPGPWizard | None: ...


@overload
def get_app_window(
    name: Literal["PasswordDialog"],
    account: str | None = None,
    jid: str | JID | None = None,
) -> PasswordDialog | None: ...


@overload
def get_app_window(
    name: Literal["PEPConfig"], account: str | None = None, jid: str | JID | None = None
) -> PEPConfig | None: ...


@overload
def get_app_window(
    name: Literal["Preferences"],
    account: str | None = None,
    jid: str | JID | None = None,
) -> Preferences | None: ...


@overload
def get_app_window(
    name: Literal["ProfileWindow"],
    account: str | None = None,
    jid: str | JID | None = None,
) -> ProfileWindow | None: ...


@overload
def get_app_window(
    name: Literal["QuitDialog"],
    account: str | None = None,
    jid: str | JID | None = None,
) -> QuitDialog | None: ...


@overload
def get_app_window(
    name: Literal["RemoveAccount"],
    account: str | None = None,
    jid: str | JID | None = None,
) -> RemoveAccount | None: ...


@overload
def get_app_window(
    name: Literal["RosterItemExchange"],
    account: str | None = None,
    jid: str | JID | None = None,
) -> RosterItemExchange | None: ...


@overload
def get_app_window(
    name: Literal["ServiceDiscoveryWindow"],
    account: str | None = None,
    jid: str | JID | None = None,
) -> ServiceDiscoveryWindow | None: ...


@overload
def get_app_window(
    name: Literal["ServiceRegistration"],
    account: str | None = None,
    jid: str | JID | None = None,
) -> ServiceRegistration | None: ...


@overload
def get_app_window(
    name: Literal["SSLErrorDialog"],
    account: str | None = None,
    jid: str | JID | None = None,
) -> SSLErrorDialog | None: ...


@overload
def get_app_window(
    name: Literal["StartChatDialog"],
    account: str | None = None,
    jid: str | JID | None = None,
) -> StartChatDialog | None: ...


@overload
def get_app_window(
    name: Literal["Themes"], account: str | None = None, jid: str | JID | None = None
) -> Themes | None: ...


@overload
def get_app_window(
    name: Literal["WorkspaceDialog"],
    account: str | None = None,
    jid: str | JID | None = None,
) -> WorkspaceDialog | None: ...


def get_app_window(
    name: str, account: str | None = None, jid: str | JID | None = None
) -> GajimWindowT | None:
    for win in app.app.get_windows():
        if win.get_name() != name:
            continue

        if hasattr(win, "wrapper"):
            win = win.wrapper  # type: ignore

        if account is not None:
            if account != win.account:  # type: ignore
                continue

        if jid is not None:
            if jid != win.jid:  # type: ignore
                continue

        return win  # type: ignore

    return None


@overload
def open_window(name: Literal["AccountWizard"], **kwargs: Any) -> AccountWizard: ...
@overload
def open_window(name: Literal["AddContact"], **kwargs: Any) -> AddContact: ...
@overload
def open_window(name: Literal["AdHocCommands"], **kwargs: Any) -> AdHocCommands: ...
@overload
def open_window(name: Literal["AdvancedConfig"], **kwargs: Any) -> AdvancedConfig: ...
@overload
def open_window(name: Literal["DBMigration"], **kwargs: Any) -> DBMigration: ...
@overload
def open_window(
    name: Literal["DebugConsoleWindow"], **kwargs: Any
) -> DebugConsoleWindow: ...
@overload
def open_window(name: Literal["CallWindow"], **kwargs: Any) -> CallWindow: ...
@overload
def open_window(
    name: Literal["CertificateDialog"], **kwargs: Any
) -> CertificateDialog: ...
@overload
def open_window(name: Literal["ChangePassword"], **kwargs: Any) -> ChangePassword: ...
@overload
def open_window(name: Literal["ContactInfo"], **kwargs: Any) -> ContactInfo: ...
@overload
def open_window(
    name: Literal["CreateGroupchatWindow"], **kwargs: Any
) -> CreateGroupchatWindow: ...
@overload
def open_window(name: Literal["Features"], **kwargs: Any) -> Features: ...
@overload
def open_window(
    name: Literal["GroupchatDetails"], **kwargs: Any
) -> GroupchatDetails: ...
@overload
def open_window(name: Literal["GroupchatJoin"], **kwargs: Any) -> GroupchatJoin: ...
@overload
def open_window(name: Literal["HistoryExport"], **kwargs: Any) -> HistoryExport: ...
@overload
def open_window(
    name: Literal["HistorySyncAssistant"], **kwargs: Any
) -> HistorySyncAssistant: ...
@overload
def open_window(name: Literal["ManageProxies"], **kwargs: Any) -> ManageProxies: ...
@overload
def open_window(name: Literal["ManageSounds"], **kwargs: Any) -> ManageSounds: ...
@overload
def open_window(name: Literal["OpenPGPSecret"], **kwargs: Any) -> OpenPGPSecret: ...
@overload
def open_window(name: Literal["OpenPGPWizard"], **kwargs: Any) -> OpenPGPWizard: ...
@overload
def open_window(name: Literal["PasswordDialog"], **kwargs: Any) -> PasswordDialog: ...
@overload
def open_window(name: Literal["PEPConfig"], **kwargs: Any) -> PEPConfig: ...
@overload
def open_window(name: Literal["Preferences"], **kwargs: Any) -> Preferences: ...
@overload
def open_window(name: Literal["ProfileWindow"], **kwargs: Any) -> ProfileWindow: ...
@overload
def open_window(name: Literal["QuitDialog"], **kwargs: Any) -> QuitDialog: ...
@overload
def open_window(name: Literal["RemoveAccount"], **kwargs: Any) -> RemoveAccount: ...
@overload
def open_window(
    name: Literal["RosterItemExchange"], **kwargs: Any
) -> RosterItemExchange: ...
@overload
def open_window(
    name: Literal["ServiceDiscoveryWindow"], **kwargs: Any
) -> ServiceDiscoveryWindow: ...
@overload
def open_window(
    name: Literal["ServiceRegistration"], **kwargs: Any
) -> ServiceRegistration: ...
@overload
def open_window(name: Literal["SSLErrorDialog"], **kwargs: Any) -> SSLErrorDialog: ...
@overload
def open_window(name: Literal["StartChatDialog"], **kwargs: Any) -> StartChatDialog: ...
@overload
def open_window(name: Literal["Themes"], **kwargs: Any) -> Themes: ...
@overload
def open_window(name: Literal["WorkspaceDialog"], **kwargs: Any) -> WorkspaceDialog: ...


def open_window(name: GajimWindowNameT, **kwargs: Any) -> GajimWindowT:
    window = get_app_window(name, kwargs.get("account"), kwargs.get("jid"))
    if window is None:
        module = import_module(WINDOW_MODULES[name])
        window_cls = getattr(module, name)
        window = window_cls(**kwargs)

    window.present()
    return window



# This file is part of Gajim.
#
# SPDX-License-Identifier: GPL-3.0-only

from __future__ import annotations

from typing import Any
from typing import TYPE_CHECKING

import datetime
import logging
import math
import sys
import wave
from collections import deque
from collections.abc import Callable
from pathlib import Path

from gi.repository import GObject

from gajim.common import app
from gajim.common import configpaths
from gajim.common.i18n import _

try:
    from gi.repository import Gst
except Exception:
    if TYPE_CHECKING:
        from gi.repository import Gst

log = logging.getLogger("gajim.gtk.voice_message_recorder")

WAV_SILENCE_START_ENABLED = True

GST_ERROR_ON_START = 0
GST_ERROR_ON_RECORDING = 1
GST_ERROR_ON_STOP = 2
GST_ERROR_ON_MERGING = 3


class VoiceMessageRecorder(GObject.GObject):
    __gtype_name__ = "VoiceMessageRecorder"

    def __init__(self, error_callback: Callable[[int, str], None]) -> None:
        GObject.GObject.__init__(self)
        self._available = True
        self._error_callback = error_callback

        # Recording state
        self._new_recording = True
        self._output_file_counter = 0
        self._output_file_valid = False
        self._output_files_invalid: list[int] = []
        self._num_samples = 80
        self._samples: deque[float] = deque(
            [0.0] * self._num_samples, maxlen=self._num_samples
        )
        self._current_peak: float = 0.0
        self._noise_floor_db: float = -45.0
        self._floor_samples: list[float] = []
        self._floor_established: bool = False
        self._recent_db: deque[float] = deque(maxlen=500)
        self._rec_time = {"start": 0, "total": 0}

        app.settings.connect_signal(
            "audio_input_device", self._on_audio_input_device_changed
        )

        # Voice message storage location
        self._filetype = "m4a"
        timestamp = datetime.datetime.now().strftime("%Y%m%d-%H%M%S")
        self._file_name = f"voice-message-{timestamp}.{self._filetype}"
        self._file_path = configpaths.get_temp_dir() / self._file_name

        self._init_pipeline()

    def _init_pipeline(self) -> None:
        log.debug("Setting up pipeline")

        self._pipeline = Gst.Pipeline.new("pipeline")
        self._queue = Gst.ElementFactory.make("queue")
        audioconvert = Gst.ElementFactory.make("audioconvert")
        audioresample = Gst.ElementFactory.make("audioresample")
        audiolevel = Gst.ElementFactory.make("level")
        capsfilt = Gst.ElementFactory.make("capsfilter")
        wavenc = Gst.ElementFactory.make("wavenc")
        self._filesink = Gst.ElementFactory.make("filesink")

        pipeline_elements = [
            self._pipeline,
            self._queue,
            audioconvert,
            audioresample,
            audiolevel,
            capsfilt,
            wavenc,
            self._filesink,
        ]

        if any(element is None for element in pipeline_elements):
            raise Exception("Error while setting up pipeline")

        assert self._pipeline is not None
        assert self._queue is not None
        assert audioconvert is not None
        assert audioresample is not None
        assert audiolevel is not None
        assert capsfilt is not None
        assert wavenc is not None
        assert self._filesink is not None

        audiolevel.set_property("message", True)
        audiolevel.set_property("interval", 10 * 1_000_000)  # 10 ms in nanoseconds
        # Force S16LE so wavenc writes standard integer PCM (format 1)
        capsfilt.set_property("caps", Gst.Caps.from_string("audio/x-raw,format=S16LE"))
        self._filesink.set_property("location", str(self._file_path))
        self._filesink.set_property("async", False)

        self._pipeline.add(self._queue)
        self._pipeline.add(audioconvert)
        self._pipeline.add(audioresample)
        self._pipeline.add(audiolevel)
        self._pipeline.add(capsfilt)
        self._pipeline.add(wavenc)
        self._pipeline.add(self._filesink)

        self._queue.link(audioconvert)
        audioconvert.link(audioresample)
        audioresample.link(audiolevel)
        audiolevel.link(capsfilt)
        capsfilt.link(wavenc)
        wavenc.link(self._filesink)

        self._clock: Gst.Clock | None = None

        self._bus = self._pipeline.get_bus()
        self._bus.add_signal_watch()
        self._id = self._bus.connect("message", self._on_gst_message)

        self._audiosrc = self._create_new_audio_src()
        if self._audiosrc is not None:
            self._pipeline.add(self._audiosrc)
            self._audiosrc.link(self._queue)

        self._available = self._audiosrc is not None
        self.notify("available")

    @GObject.Property(type=bool, default=True, flags=GObject.ParamFlags.READABLE)
    def available(self) -> bool:
        return self._available

    def _set_property_available(self, value: bool) -> None:
        if value == self._available:
            return
        self._available = value
        self.notify("available")

    @property
    def recording_in_progress(self) -> bool:
        _, pipeline_state, _ = self._pipeline.get_state(timeout=100)
        return pipeline_state != Gst.State.NULL

    @property
    def audio_file_is_valid(self) -> bool:
        return self._output_file_valid

    @property
    def audio_file_abspath(self) -> Path:
        return self._file_path

    @property
    def audio_file_uri(self) -> str:
        uri = self._file_path.as_uri()
        assert uri is not None
        return uri

    def cleanup(self) -> None:
        del self._error_callback
        self._file_path.unlink(missing_ok=True)
        self._pipeline.get_bus().disconnect(self._id)

    def start_recording(self) -> None:
        assert self._filesink is not None

        if self._new_recording:
            self._new_recording = False
            timestamp = datetime.datetime.now().strftime("%Y%m%d-%H%M%S")
            self._file_name = f"voice-message-{timestamp}.{self._filetype}"
            self._file_path = configpaths.get_temp_dir() / self._file_name

        if not self._file_path.parent.exists():
            self._handle_error_output_dir_inaccessible()
            return

        self._output_file_counter += 1
        self._filesink.set_property(
            "location", f"{self._file_path}.part{self._output_file_counter}"
        )
        log.debug(
            "Creating new recording file: %s", self._filesink.get_property("location")
        )
        self._pipeline.set_state(Gst.State.PLAYING)
        message = self._pipeline.get_bus().pop_filtered(Gst.MessageType.ERROR)
        if message is not None:
            self._handle_error_on_start(message)
        self._clock = self._pipeline.get_pipeline_clock()
        self._rec_time["start"] = self._clock.get_time()

    def stop_recording(self) -> None:
        if not self.recording_in_progress:
            return

        assert self._clock is not None

        self._rec_time["total"] += self._clock.get_time() - self._rec_time["start"]

        self._pipeline.send_event(Gst.Event.new_eos())
        message = self._pipeline.get_bus().timed_pop_filtered(
            Gst.CLOCK_TIME_NONE, Gst.MessageType.EOS | Gst.MessageType.ERROR
        )
        self._pipeline.set_state(Gst.State.NULL)
        assert message is not None
        self._handle_error_on_stop(message)

        self._merge_output_files()

    def stop_and_reset(self) -> None:
        if self.recording_in_progress:
            self.stop_recording()

        self._new_recording = True
        self._output_file_valid = False
        self._output_file_counter = 0
        self._output_files_invalid = []
        self._samples = deque([0] * self._num_samples, maxlen=self._num_samples)
        self._current_peak = 0.0
        self._noise_floor_db = -45.0
        self._floor_samples = []
        self._floor_established = False
        self._recent_db.clear()
        self._rec_time = {"start": 0, "total": 0}

    def cancel_recording(self) -> None:
        # Immediately abort without draining EOS or merging files, so short taps
        # that triggered an early pipeline start don't block the main thread.
        self._pipeline.set_state(Gst.State.NULL)
        self._new_recording = True
        self._output_file_valid = False
        self._output_file_counter = 0
        self._output_files_invalid = []
        self._samples = deque([0] * self._num_samples, maxlen=self._num_samples)
        self._current_peak = 0.0
        self._noise_floor_db = -45.0
        self._floor_samples = []
        self._floor_established = False
        self._recent_db.clear()
        self._rec_time = {"start": 0, "total": 0}

    def recording_time(self) -> int:
        delta = 0
        if self.recording_in_progress and self._clock is not None:
            delta = self._clock.get_time() - self._rec_time["start"]
        return self._rec_time["total"] + delta

    def request_new_sample(self) -> None:
        peak = self._current_peak
        self._current_peak = 0.0
        prev = self._samples[0] if self._samples else 0.0
        # Release envelope: decay rather than hard-cut to zero on silence
        released = prev * 0.5
        self._samples.appendleft(max(peak, released))

    def _normalize_peak_db(self, peak_db: float) -> float:
        self._recent_db.append(peak_db)

        if not self._floor_established:
            self._floor_samples.append(peak_db)
            if len(self._floor_samples) >= 5:  # 5 × 10 ms = 50 ms
                self._noise_floor_db = sum(self._floor_samples) / len(
                    self._floor_samples
                )
                self._floor_established = True
            return 0.0
        elif len(self._recent_db) >= 50:
            sorted_levels = sorted(self._recent_db)
            idx = int(len(sorted_levels) * 0.10)
            floor_estimate = sorted_levels[idx]
            # Very slow EMA so the floor never visibly shifts during recording
            self._noise_floor_db = self._noise_floor_db * 0.999 + floor_estimate * 0.001

        adjusted = peak_db - self._noise_floor_db - 3.0
        if adjusted <= 0.0:
            return 0.0

        value = min(1.0, adjusted / 50.0)
        return math.sqrt(value)

    def recording_samples(self) -> list[tuple[float, float]]:
        sample_list = reversed(list(self._samples))
        result: list[tuple[float, float]] = []
        for sample in sample_list:
            result.append((sample, sample))
        return result

    def _is_error(self, message: Gst.Message | None) -> bool:
        if message is not None and message.type == Gst.MessageType.ERROR:
            self._log_error(message)
            return True
        return False

    def _log_error(self, message: Gst.Message | None) -> None:
        if message is None:
            return

        structure = message.get_structure()
        if structure is not None:
            gerror = structure.get_value("gerror")
            debug = structure.get_value("debug")

            assert gerror is not None
            log.error("gerror code: %s", gerror.code)
            log.error("gerror domain: %s", gerror.domain)
            log.error("debug: %s", debug)

    def _error_info(self, message: Gst.Message | None) -> tuple[str | None, int | None]:
        if message is None:
            return None, None
        structure = message.get_structure()
        assert structure is not None
        gerror = structure.get_value("gerror")
        assert gerror is not None
        return gerror.domain, gerror.code

    def _handle_error_on_start(self, message: Gst.Message) -> None:
        assert message is not None
        if not self._is_error(message):
            return

        log.error("Error on starting the recording.")
        self._output_files_invalid.append(self._output_file_counter)

        domain, code = self._error_info(message)
        if domain == "gst-resource-error-quark":
            if code == 5:
                # GST_RESOURCE_ERROR_OPEN_READ (5)
                # used when resource fails to open for reading.
                self._error_callback(
                    GST_ERROR_ON_START, _("Is a microphone plugged in and accessible?")
                )
        if domain == "gst-stream-error-quark":
            if code == 1:
                # GST_STREAM_ERROR_FAILED (1) also fires when no audio device is available.
                self._error_callback(
                    GST_ERROR_ON_START, _("Is a microphone plugged in and accessible?")
                )

    def _handle_error_on_recording(self, message: Gst.Message | None):
        if message is not None and not self._is_error(message):
            return

        log.error("Error during the recording.")
        self._output_files_invalid.append(self._output_file_counter)

        domain, code = self._error_info(message)
        if domain == "gst-resource-error-quark":
            if code == 9:
                # GST_RESOURCE_ERROR_READ (9)
                # used when the resource can't be read from.
                self._error_callback(
                    GST_ERROR_ON_RECORDING,
                    _("Is a microphone plugged in and accessible?"),
                )

        if domain == "gst-stream-error-quark":
            if code == 1:
                # GST_STREAM_ERROR_FAILED (1) also fires when no audio device is available.
                self._error_callback(
                    GST_ERROR_ON_RECORDING,
                    _("Is a microphone plugged in and accessible?"),
                )

    def _handle_error_on_stop(self, message: Gst.Message | None) -> None:
        if message is not None and not self._is_error(message):
            return
        # TODO
        log.error("Error when stopping the recording.")

    def _handle_error_on_merging(self, message: Gst.Message | None) -> None:
        if message is not None and not self._is_error(message):
            return

        # TODO
        log.error("Error when merging the recordings.")

    def _handle_error_output_dir_inaccessible(self) -> None:
        log.error(
            "Voice message lost. Temporary output folder %s not accessible",
            self._file_path.parent,
        )
        error_message = _("Voice message could not be saved. Please try again.")
        self._error_callback(GST_ERROR_ON_MERGING, error_message)
        self.stop_and_reset()

    def _on_gst_message(self, _bus: Gst.Bus, message: Gst.Message) -> None:
        structure = message.get_structure()
        if structure is None:
            return
        message_string = structure.to_string()

        if message.type == Gst.MessageType.ELEMENT:
            name = structure.get_name()
            if name == "level":
                if not structure.has_field("peak"):
                    return

                peak_values = structure.get_value("peak")
                assert peak_values is not None
                if len(peak_values) > 0:
                    peak_db = float(peak_values[0])
                    if not math.isinf(peak_db) and peak_db > -90.0:
                        normalized = self._normalize_peak_db(peak_db)
                        self._current_peak = max(self._current_peak, normalized)
            log.debug("gst element message: %s", message_string)

        elif message.type == Gst.MessageType.ERROR:
            self._handle_error_on_recording(message)

        elif message.type == Gst.MessageType.EOS:
            pass

    def _switch_sources(self) -> None:
        assert self._queue is not None
        assert not self.recording_in_progress

        if self._audiosrc is not None:
            self._audiosrc.unlink(self._queue)
            self._pipeline.remove(self._audiosrc)

        self._audiosrc = self._create_new_audio_src()
        if self._audiosrc is not None:
            self._pipeline.add(self._audiosrc)
            self._audiosrc.link(self._queue)

        self._available = self._audiosrc is not None
        self.notify("available")

    def _create_new_audio_src(self) -> Gst.Element | None:
        device = self._extract_first_word(app.settings.get("audio_input_device"))

        if sys.platform == "win32" and device == "autoaudiosrc":
            device = "wasapisrc"

        if device == "audiotestsrc":
            log.error('Audio device "audiotestsrc" not supported')
            return None

        audiosrc = Gst.ElementFactory.make(device)
        if audiosrc is None:
            log.debug("Unable to create audio source for: %s", device)
        else:
            log.debug("Created audio source for: %s", device)
            if device == "wasapisrc":
                audiosrc.set_property("role", "comms")

        return audiosrc

    def _on_audio_input_device_changed(self, device: str, *args: Any) -> None:
        log.debug("Switching to input device: %s", device)
        self._switch_sources()

    def _file_merge_required(self) -> bool:
        return self._output_file_counter > 1

    def _smooth_wav_boundaries(
        self, file_path: Path, silence_ms: int = 20, fade_ms: int = 30
    ) -> None:
        try:
            with wave.open(str(file_path), "rb") as wav_in:
                params = wav_in.getparams()
                frames = wav_in.readframes(params.nframes)
        except Exception as e:
            log.warning("Skipping boundary smoothing for %s: %s", file_path, e)
            return

        n_channels = params.nchannels
        sampwidth = params.sampwidth
        n_frames = params.nframes
        framerate = params.framerate

        silence_frames = min(int(framerate * silence_ms / 1000), n_frames)
        fade_frames = min(int(framerate * fade_ms / 1000), n_frames - silence_frames)

        buf = bytearray(frames)

        # Zero the initial silence region
        silence_bytes = silence_frames * n_channels * sampwidth
        buf[:silence_bytes] = b"\x00" * silence_bytes

        if sampwidth == 2:
            # Fade-in after the silence to avoid a step discontinuity at the start
            for i in range(fade_frames):
                factor = i / fade_frames
                frame_start = (silence_frames + i) * n_channels * sampwidth
                for ch in range(n_channels):
                    offset = frame_start + ch * sampwidth
                    value = int.from_bytes(
                        buf[offset : offset + 2], "little", signed=True
                    )
                    buf[offset : offset + 2] = int(value * factor).to_bytes(
                        2, "little", signed=True
                    )

            # Fade-out at the end to avoid a step discontinuity at the merge point
            fade_out_frames = min(int(framerate * fade_ms / 1000), n_frames)
            for i in range(fade_out_frames):
                factor = i / fade_out_frames
                frame_start = (n_frames - 1 - i) * n_channels * sampwidth
                for ch in range(n_channels):
                    offset = frame_start + ch * sampwidth
                    value = int.from_bytes(
                        buf[offset : offset + 2], "little", signed=True
                    )
                    buf[offset : offset + 2] = int(value * factor).to_bytes(
                        2, "little", signed=True
                    )

        with wave.open(str(file_path), "wb") as wav_out:
            wav_out.setparams(params)
            wav_out.writeframes(bytes(buf))

    def _build_merge_command(self) -> str:
        log.info("Merging WAV files started")

        # Use as_posix() on file path to convert "\" to "/" on Windows
        sources = ""
        for i in range(1, self._output_file_counter + 1):
            if i in self._output_files_invalid:
                continue
            source = " ! ".join(
                [
                    f"filesrc location={self._file_path.as_posix()}.part{i}",
                    "wavparse",
                    "audioconvert",
                    "audioresample",
                    "c. ",
                ]
            )
            sources += source

        command = " ! ".join(
            [
                "concat name=c",
                "queue",
                "audioconvert",
                "audioresample",
                "opusenc audio-type=voice bitrate=32000",
                "mp4mux ",
                f"filesink location={self._file_path.as_posix()} {sources}",
            ]
        )
        return command

    def _merge_output_files(self) -> None:
        if not self._file_path.parent.exists():
            self._handle_error_output_dir_inaccessible()
            return

        for i in range(1, self._output_file_counter + 1):
            if i in self._output_files_invalid:
                continue
            part_path = Path(f"{self._file_path}.part{i}")
            if WAV_SILENCE_START_ENABLED:
                self._smooth_wav_boundaries(part_path)

        command = self._build_merge_command()
        pipeline = Gst.parse_launch(command)

        # Remove the original output file first before writing into it
        if self._file_path.exists():
            self._file_path.unlink()

        pipeline.set_state(Gst.State.PLAYING)

        bus = pipeline.get_bus()
        if bus is None:
            return

        message = bus.timed_pop_filtered(
            Gst.CLOCK_TIME_NONE, Gst.MessageType.EOS | Gst.MessageType.ERROR
        )
        pipeline.set_state(Gst.State.NULL)

        self._handle_error_on_merging(message)
        self._output_file_valid = True
        log.debug("Merging files finished")

    @staticmethod
    def _extract_first_word(gst_cmd: str) -> str:
        """
        Gajim gives us a string with an audio src plus additional parameters
        such as volume. This method takes only the first word == audio src.
        """
        if " " in gst_cmd:
            idx = gst_cmd.index(" ")
            gst_cmd = gst_cmd[:idx]
        return gst_cmd



# This file is part of Gajim.
#
# SPDX-License-Identifier: GPL-3.0-only

from gi.repository import Gio
from gi.repository import Gtk

from gajim.common import app
from gajim.common.client import Client
from gajim.common.const import GIO_TLS_ERRORS
from gajim.common.i18n import _

from gajim.gtk.builder import get_builder
from gajim.gtk.util.window import open_window
from gajim.gtk.window import GajimAppWindow


class SSLErrorDialog(GajimAppWindow):
    def __init__(
        self,
        account: str,
        client: Client,
        cert: Gio.TlsCertificate,
        ignored_errors: set[Gio.TlsCertificateFlags],
        error: Gio.TlsCertificateFlags,
    ) -> None:
        GajimAppWindow.__init__(
            self,
            name="SSLErrorDialog",
            title=_("SSL Certificate Verification Error"),
            add_window_padding=True,
            header_bar=True,
        )

        self._ui = get_builder("ssl_error_dialog.ui")
        self.set_child(self._ui.ssl_error_box)

        self._connect(
            self._ui.add_certificate_checkbutton,
            "toggled",
            self._on_add_certificate_toggled,
        )
        self._connect(self._ui.view_cert_button, "clicked", self._on_view_cert_clicked)
        self._connect(self._ui.connect_button, "clicked", self._on_connect_clicked)

        self.account = account
        self._error = error
        self._ignored_errors = ignored_errors
        self._client = client
        self._cert = cert
        self._server = app.get_hostname_from_account(self.account)

        self._process_error()

    def _cleanup(self) -> None:
        pass

    def _process_error(self) -> None:
        self._ui.intro_text.set_text(
            _(
                "There was an error while attempting to verify the SSL "
                "certificate of your XMPP server (%s)."
            )
            % self._server
        )

        unknown_error = _('Unknown SSL error "%s"') % self._error
        ssl_error = GIO_TLS_ERRORS.get(self._error, unknown_error)
        self._ui.ssl_error.set_text(ssl_error)

        if self._error == Gio.TlsCertificateFlags.UNKNOWN_CA:
            self._ui.add_certificate_checkbutton.set_visible(True)

        elif self._error == Gio.TlsCertificateFlags.EXPIRED:
            self._ui.connect_button.set_sensitive(True)

        else:
            self._ui.connect_button.set_visible(False)
            self._ui.connect_button.set_visible(False)

    def _on_view_cert_clicked(self, _button: Gtk.Button) -> None:
        open_window(
            "CertificateDialog",
            account=self.account,
            transient_for=self,
            cert=self._cert,
        )

    def _on_add_certificate_toggled(self, checkbutton: Gtk.CheckButton) -> None:
        self._ui.connect_button.set_sensitive(checkbutton.get_active())

    def _on_connect_clicked(self, _button: Gtk.Button) -> None:
        if self._ui.add_certificate_checkbutton.get_active():
            app.cert_store.add_certificate(self._cert)

        if self._error == Gio.TlsCertificateFlags.EXPIRED:
            self._ignored_errors.add(Gio.TlsCertificateFlags.EXPIRED)

        self._client.connect(ignored_tls_errors=self._ignored_errors)
        self.close()
